#---------------------------------------------------------------------------
# configuration file
#---------------------------------------------------------------------------

############################################################################
# handy class for dictionary
class Dict(dict):
    def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs)
    def __getattr__(self, name): return self[name]
    def __setattr__(self, name, value): self[name] = value
    def __delattr__(self, name): del self[name]

############################################################################

#---------------------------------------------------------------------------
# Dataset builder-parameters
# EDIT THE FOLLOWING PARAMATERS WITH YOUR OWN DATASETS PARAMETERS

# Folder where pre-processed images are stored
IMG_DIR = 'data/pancreas/tif_imagesTr_small'

# Folder where pre-processed masks are stored
MSK_DIR = 'data/pancreas/tif_labelsTr_small'

# (optional) path to the .csv file storing "filename,hold_out,fold", where:
# "filename" is the image name,
# "hold_out" is either 0 (training image) or 1 (testing image),
# "fold" (non-negative integer) indicates the k-th fold, 
# by default fold 0 of the training image (hold_out=0) is the validation set.
# CSV_DIR = 'data/pancreas/folds_pancreas.csv'

# CSV_DIR can be set to None, in which case the validation set will be
# automatically chosen from the training set (20% of the training images/masks)
CSV_DIR = None 

# model name
DESC = 'unet_default'

# number of classes of objects
# the background does not count, so the minimum is 1 (the max is 255)
NUM_CLASSES=2

#---------------------------------------------------------------------------
# Auto-config builder-parameters
# PASTE AUTO-CONFIG RESULTS HERE

# batch size
BATCH_SIZE = 2

# patch size passed to the model
PATCH_SIZE = [40,224,224]

# larger patch size used prior rotation augmentation to avoid "empty" corners.
AUG_PATCH_SIZE = [48,263,263]

# number of pooling done in the UNet
NUM_POOLS = [3,5,5]

# median spacing is used only during prediction to normalize the output images
# it is commented here because we did not noticed any improvemet
# MEDIAN_SPACING=[0.79492199, 0.79492199, 2.5]
MEDIAN_SPACING=[]

#---------------------------------------------------------------------------
# Advanced paramaters (can be left as such) 
# training configs

# whether to store also the best model 
SAVE_BEST = True 

# number of epochs
# the number of epochs can be reduced for small training set (e.g. a set of 10 images/masks of 128x128x64)
NB_EPOCHS = 1000

# optimizer paramaters
LR_START = 1e-2 # comment if need to reload learning rate after training interruption
WEIGHT_DECAY = 3e-5

# whether to use deep-supervision loss:
# a loss is placed at each stage of the UNet model
USE_DEEP_SUPERVISION = False

# whether to use softmax loss instead of sigmoid
# should not be set to True if object classes are overlapping in the masks
USE_SOFTMAX=False 

# training loop parameters
USE_FP16 = True
NUM_WORKERS = 4

#---------------------------------------------------------------------------
# callback setup (can be left as such) 
# callbacks are routines that execute periodically during the training loop

# folder where the training logs will be stored, including:
# - model .pth files (state_dict)
# - image snapshots of model training (only if USE_IMAGE_CLBK is True)
# - logs with this configuration stored in .yaml format and tensorboard logs
LOG_DIR = 'logs/'

SAVE_MODEL_EVERY_EPOCH = 1
USE_IMAGE_CLBK = True
VAL_EVERY_EPOCH = SAVE_MODEL_EVERY_EPOCH
SAVE_IMAGE_EVERY_EPOCH = SAVE_MODEL_EVERY_EPOCH
USE_FG_CLBK = True

#---------------------------------------------------------------------------
# dataset configs

TRAIN_DATASET = Dict(
    fct="SegPatchFast",
    kwargs=Dict(
        img_dir    = IMG_DIR,
        msk_dir    = MSK_DIR, 
        batch_size = BATCH_SIZE, 
        patch_size = PATCH_SIZE,
        nbof_steps = 250,
        folds_csv  = CSV_DIR, 
        fold       = 0, 
        val_split  = 0.20,
        train      = True,
        use_aug    = True,
        aug_patch_size = AUG_PATCH_SIZE,
        use_softmax  = USE_SOFTMAX,
    )
)

TRAIN_DATALOADER_KWARGS = Dict(
    batch_size  =BATCH_SIZE, 
    drop_last   =True, 
    shuffle     =True, 
    num_workers =NUM_WORKERS, 
    pin_memory  =True,
)          

VAL_DATASET = Dict(
    fct="SegPatchFast",
    kwargs = Dict(
        img_dir    = IMG_DIR,
        msk_dir    = MSK_DIR, 
        batch_size = BATCH_SIZE, 
        patch_size = PATCH_SIZE,
        nbof_steps = 50,
        folds_csv  = CSV_DIR, 
        fold       = 0, 
        val_split  = 0.20,
        train      = False,
        use_aug    = False,
        use_softmax  = USE_SOFTMAX,
        fg_rate    = 0.33,
    )
)

VAL_DATALOADER_KWARGS = Dict(
    batch_size  =BATCH_SIZE, # TODO: change it in the final version
    drop_last   =False, 
    shuffle     =False, 
    num_workers =NUM_WORKERS, 
    pin_memory  =True,
)

#---------------------------------------------------------------------------
# model configs

MODEL = Dict(
    fct="UNet3DVGGDeep", # from the register
    kwargs = Dict(
        num_pools=NUM_POOLS,
        num_classes=NUM_CLASSES,
        factor = 32,
        use_deep=USE_DEEP_SUPERVISION,
    )
)

#---------------------------------------------------------------------------
# loss configs

TRAIN_LOSS = Dict(
    fct="DiceBCE",
    kwargs = Dict(name="train_loss", use_softmax=USE_SOFTMAX)
)

VAL_LOSS = Dict(
    fct="DiceBCE",
    kwargs = Dict(name="val_loss", use_softmax=USE_SOFTMAX)
)

#---------------------------------------------------------------------------
# metrics configs


VAL_METRICS = Dict(
    val_iou=Dict(
        fct="IoU",
        kwargs = Dict(name="val_iou", use_softmax=USE_SOFTMAX)),
    val_dice=Dict(
        fct="Dice",
        kwargs=Dict(name="val_dice", use_softmax=USE_SOFTMAX)),
)

#---------------------------------------------------------------------------
# trainers configs

TRAINER = Dict(
    fct="SegTrain",
    kwargs=Dict(),
)

VALIDATER = Dict(
    fct="SegVal",
    kwargs=Dict(),
)

#---------------------------------------------------------------------------
# predictors configs

PREDICTOR = Dict(
    fct="SegPatch",
    kwargs=Dict(patch_size=PATCH_SIZE, tta=True, median_spacing=MEDIAN_SPACING, use_softmax=USE_SOFTMAX, keep_biggest_only=False),
)

############################################################################
# end of config file
# do not write anything in or below this field

CONFIG = Dict(**globals().copy()) # stores all variables in one Dict

to_remove = ['__name__', '__doc__', '__package__',
    '__loader__' , '__spec__', '__annotations__',
    '__builtins__', '__file__', '__cached__', 'Dict']

for k in to_remove: 
    if (k in CONFIG.keys()): del CONFIG[k] 

############################################################################



from skimage.io import imread
import os
import numpy as np
import argparse

def abs_path(root, listdir_):
    listdir = listdir_.copy()
    for i in range(len(listdir)):
        listdir[i] = root + '/' + listdir[i]
    return listdir

def abs_listdir(path):
    return abs_path(path, os.listdir(path))

def compute_median(path):
    path_imgs = abs_listdir(path)
    sizes = []
    for i in range(len(path_imgs)):
        sizes += [list(imread(path_imgs[i]).shape)]
    sizes = np.array(sizes)
    median = np.median(sizes, axis=0).astype(int)
    # print("median:",median)
    return median 

def single_patch_pool(dim, size_limit=7):
    """
    divide by two the dim number until obtaining a number lower than 7
    then np.ceil this number
    then multiply multiple times by two this number to obtain the patch size
    """
    pool = 0
    while dim > size_limit:
        dim /= 2
        pool += 1
    patch = np.round(dim)
    patch = patch.astype(int)*(2**pool)
    return patch, pool

def find_patch_pool_batch(dims, max_dims=(128,128,128), max_pool=5, epsilon=1e-3):
    """
    take the median size as input, determine the patch size and the number of pool
    with "single_patch_pool" function for each dimension and 
    assert that the final dimension size is smaller than max_dims.prod().
    """
    # transform tuples into arrays
    assert len(dims)==3 or len(dims)==4
    if len(dims)==4:
        dims=dims[1:]
    dims = np.array(dims)
    max_dims = np.array(max_dims)
    
    # divides by a 1+epsilon until reaching a sufficiently small resolution
    while dims.prod() > max_dims.prod():
        dims = dims / (1+epsilon)
    dims = dims.astype(int)
    
    # compute patch and pool for all dims
    patch_pool = np.array([single_patch_pool(m) for m in dims])
    patch = patch_pool[:,0]
    pool = patch_pool[:,1]
    
    # assert the final size is smaller than max_dims
    while patch.prod()>max_dims.prod():
        patch = patch - np.array([32,32,32])*(patch>max_dims) # removing multiples of 32
    pool = np.where(pool > max_pool, max_pool, pool)
    
    # batch_size
    batch = 2
    while batch*patch.prod() <= 2*max_dims.prod():
        batch += 1
    if batch*patch.prod() > 2*max_dims.prod():
        batch -= 1
    return patch, pool, batch

def display_info(patch, pool, batch):
    print("*"*20,"YOU CAN COPY AND PASTE THE FOLLOWING LINES INSIDE THE CONFIG FILE", "*"*20)
    print("BATCH_SIZE =", batch)
    print("PATCH_SIZE =", list(patch))
    aug_patch = np.array(patch)+2**(np.array(pool)+1)
    print("AUG_PATCH_SIZE =",list(aug_patch))
    print("NUM_POOLS =", list(pool))

def auto_config(img_dir, max_dims=(128,128,128)):
    median = compute_median(path=img_dir)
    patch, pool, batch = find_patch_pool_batch(dims=median, max_dims=max_dims) 
    aug_patch = np.array(patch)+2**(np.array(pool)+1)
    return batch, aug_patch, patch, pool

def minimal_display(img_dir, max_dims=(128,128,128)):
    out = auto_config(img_dir, max_dims=max_dims)
    for element in out:
        print(element)

if __name__=='__main__':
    parser = argparse.ArgumentParser(description="Dataset preprocessing for training purpose.")
    parser.add_argument("--img_dir", type=str,
        help="Path of the images directory")
    parser.add_argument("--max_dim", type=int, default=128,
        help="Maximum size of one dimension of the patch (default: 128)")  
    parser.add_argument("--min_dis", default=False,  action='store_true', dest='min_dis',
        help="Minimal display. Display only the raw batch, aug_patch, patch and pool")
    args = parser.parse_args()


    if args.min_dis:
        minimal_display(img_dir=args.img_dir, max_dims=(args.max_dim, args.max_dim, args.max_dim))
    else:
        median = compute_median(path=args.img_dir)
        patch, pool, batch = find_patch_pool_batch(dims=median, max_dims=(args.max_dim, args.max_dim, args.max_dim))
        display_info(patch, pool, batch)

    # median=compute_median(path='/home/gumougeot/all/codes/python/3dnucleus/data/pancreas/tif_imagesTr_small')
    # median=compute_median(path='/home/gumougeot/all/codes/python/3dnucleus/data/lung/tif_imagesTr')
    # median=compute_median(path='/home/gumougeot/all/codes/python/3dnucleus/data/remi/tif_img')
    # print("patch, pool, batch:", find_patch_pool_batch(dims=median, max_dims=(160,160,160)))
    # print("patch, pool, batch:", find_patch_pool_batch(dims=median, max_dims=(192,192,192)))
    # print("patch, pool, batch:", find_patch_pool_batch(dims=median, max_dims=(128,128,128)))




#---------------------------------------------------------------------------
# configuration file
#---------------------------------------------------------------------------

############################################################################
# handy class for dictionary

class Dict(dict):
    def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs)
    def __getattr__(self, name): return self[name]
    def __setattr__(self, name, value): self[name] = value
    def __delattr__(self, name): del self[name]

############################################################################

#---------------------------------------------------------------------------
# Logs configs

# CSV_DIR = 'data/pancreas/folds_pancreas.csv'
# CSV_DIR = 'data/nucleus/folds_sophie.csv'
CSV_DIR = None

IMG_DIR = 'data/chromo/tif_img'
# IMG_DIR = 'data/pancreas/tif_imagesTr_small'
# IMG_DIR = 'data/pancreas/tif_train_img'
# IMG_DIR = 'data/pancreas/tif_train_img_17'
# IMG_DIR = 'data/aline/img_resampled'
# IMG_DIR = 'data/nucleus/images_resampled'
# IMG_DIR = 'data/nucleus/images_resampled_sophie'
# IMG_DIR = 'data/nucleus/images_manual'

MSK_DIR = 'data/chromo/tif_msk'
# MSK_DIR = 'data/pancreas/tif_labelsTr_small'
# MSK_DIR = 'data/pancreas/tif_train_img_labels'
# MSK_DIR = 'data/pancreas/tif_train_img_labels_17'
# MSK_DIR = 'data/aline/msk_resampled'
# MSK_DIR = 'data/nucleus/masks_resampled_sophie'
# MSK_DIR = 'data/nucleus/masks_resampled'
# MSK_DIR = 'data/nucleus/masks_manual'

# PT_PATH = 'data/pancreas/data_pancreas.pt'
LOG_DIR = 'logs/'

# DESC = 'unet_mine-pancreas_21'
DESC = 'unet_mine-chromo'

#---------------------------------------------------------------------------
# training configs

SAVE_BEST = True # whether we save also the best model 

NB_EPOCHS = 100
BATCH_SIZE = 8
LR_START = 1e-2 # comment if need to reload learning rate after training interruption
# LR_MILESTONES = [100, NB_EPOCHS//2, NB_EPOCHS-100]
# LR_T_MAX = NB_EPOCHS
WEIGHT_DECAY = 3e-5

USE_DEEP_SUPERVISION = False
# NUM_POOLS = [3,5,5]
NUM_POOLS = [3,4,4]
# NUM_POOLS = [4,5,5]
# NUM_CLASSES=3
NUM_CLASSES=1
USE_SOFTMAX=False 

# PATCH_SIZE = [40,224,224]
# PATCH_SIZE = [48,256,256]
# PATCH_SIZE = [64,320,320]
# PATCH_SIZE = [53,266,266]
PATCH_SIZE = [56,64,64]

# AUG_PATCH_SIZE = [48,263,263]
# AUG_PATCH_SIZE = [48,256,256]
AUG_PATCH_SIZE = [72,96,96]

# MEDIAN_SPACING=[0.79492199, 0.79492199, 2.5]
MEDIAN_SPACING=[]

USE_FP16 = True

#---------------------------------------------------------------------------
# callback setup

SAVE_MODEL_EVERY_EPOCH = 1
USE_IMAGE_CLBK = True
VAL_EVERY_EPOCH = SAVE_MODEL_EVERY_EPOCH
SAVE_IMAGE_EVERY_EPOCH = SAVE_MODEL_EVERY_EPOCH
USE_FG_CLBK = True

#---------------------------------------------------------------------------
# dataset configs

# TRAIN_DATASET = Dict(
#     fct="Seg",
#     kwargs=Dict(
#         img_dir     =IMG_DIR,
#         msk_dir     =MSK_DIR,  
#         folds_csv   =CSV_DIR,
#         fold        = 0,
#         val_split   = 0.36,
#         train       =True,
#         use_onehot  =False,
#     )
# )



# TRAIN_DATASET = Dict(
#     fct="SegFast",
#     kwargs=Dict(
#         pt_path     =PT_PATH,
#         val_split   = 0.2,
#         train       =True,
#     )
# )

TRAIN_DATASET = Dict(
    fct="SegPatchFast",
    kwargs=Dict(
        img_dir    = IMG_DIR,
        msk_dir    = MSK_DIR, 
        batch_size = BATCH_SIZE, 
        patch_size = PATCH_SIZE,
        nbof_steps = 250,
        folds_csv  = CSV_DIR, 
        fold       = 0, 
        val_split  = 0.20,
        train      = True,
        use_aug    = True,
        aug_patch_size = AUG_PATCH_SIZE,
        use_softmax  = USE_SOFTMAX,
    )
)

TRAIN_DATALOADER_KWARGS = Dict(
    batch_size  =BATCH_SIZE, 
    drop_last   =True, 
    shuffle     =True, 
    num_workers =6, 
    # num_workers =0,
    pin_memory  =True,
)          


# TRAIN_DATALOADER = Dict(
#     fct="SegPatchFast",
#     kwargs = Dict(
#         img_dir    = IMG_DIR,
#         msk_dir    = MSK_DIR, 
#         batch_size = BATCH_SIZE, 
#         patch_size = [40,224,224],
#         nbof_steps = 250,
#         folds_csv  = CSV_DIR, 
#         fold       = 0, 
#         val_split  = 0.25,
#         train      = True,
#         use_aug    = False,
#     )
# )


# VAL_DATASET = Dict(
#     fct="Seg",
#     kwargs = Dict(
#         img_dir     =IMG_DIR,
#         msk_dir     =MSK_DIR,  
#         folds_csv   =CSV_DIR,
#         fold        = 0,
#         val_split   = 0.36,
#         train       =False,
#         use_onehot  =False,
#     )
# )

# VAL_DATASET = Dict(
#     fct="SegFast",
#     kwargs = Dict(
#         pt_path     =PT_PATH,
#         val_split   = 0.2,
#         train       =False,
#     )
# )

VAL_DATASET = Dict(
    fct="SegPatchFast",
    kwargs = Dict(
        img_dir    = IMG_DIR,
        msk_dir    = MSK_DIR, 
        batch_size = BATCH_SIZE, 
        patch_size = PATCH_SIZE,
        nbof_steps = 50,
        folds_csv  = CSV_DIR, 
        fold       = 0, 
        val_split  = 0.20,
        train      = False,
        use_aug    = False,
        use_softmax  = USE_SOFTMAX,
        fg_rate    = 0.33,
    )
)

VAL_DATALOADER_KWARGS = Dict(
    batch_size  =2, # TODO: change it in the final version
    drop_last   =False, 
    shuffle     =False, 
    num_workers =6, 
    # num_workers =0, 
    pin_memory  =True,
)


# VAL_DATALOADER = Dict(
#     fct="SegPatchFast",
#     kwargs = Dict(
#         img_dir    = IMG_DIR,
#         msk_dir    = MSK_DIR, 
#         batch_size = BATCH_SIZE, 
#         patch_size = [40,224,224],
#         nbof_steps = 50,
#         folds_csv  = CSV_DIR, 
#         fold       = 0, 
#         val_split  = 0.25,
#         train      = False,
#         use_aug    = False,
#     )
# )

#---------------------------------------------------------------------------
# model configs

MODEL = Dict(
    fct="UNet3DVGGDeep", # from the register
    kwargs = Dict(
        num_pools=NUM_POOLS,
        num_classes=NUM_CLASSES,
        factor = 32,
        use_deep=USE_DEEP_SUPERVISION,
    )
)



# MODEL = Dict(
#     fct="FPN", # from the register
#     kwargs = Dict(
#         # pyramid={                       # size  fmaps
#         # 0: ['_bn0'],                    # 64    32
#         # 1: ['_blocks', '0', '_bn2'],    # 32    16
#         # 2: ['_blocks', '2', '_bn2'],    # 16    24
#         # 3: ['_blocks', '4', '_bn2'],    # 8     40
#         # 4: ['_blocks', '10', '_bn2'],   # 4     112
#         # 5: ['_blocks', '15', '_bn2']    # 2     320
#         # },
#         pyramid={                       # efficientnet b4
#         0: ['_bn0'],                    
#         1: ['_blocks', '1', '_bn2'],   
#         2: ['_blocks', '5', '_bn2'],  
#         3: ['_blocks', '9', '_bn2'],    
#         4: ['_blocks', '21', '_bn2'],  
#         # 5: ['_blocks', '31', '_bn2']  
#         },
#         in_dim=1,
#         out_dim=1,
#         num_filters=1,
#     )
# )


# MODEL = Dict(
#     fct="BasicUNetMonai",
#     kwargs = Dict(out_channels=2),
# )

#---------------------------------------------------------------------------
# loss configs

# TRAIN_LOSS = Dict(
#     fct="DeepDiceBCE",
#     kwargs = Dict(
#         milestones=[200],
#         alphas=[1/5, 0],
#         name="train_loss",
#     )
# )

# TRAIN_LOSS = Dict(
#     fct="DeepDiceBCE",
#     kwargs = Dict(
#         alphas=[0, 0, 0.06667, 0.1333, 0.2667, 0.5333],
#         name="train_loss",
#     )
# )

TRAIN_LOSS = Dict(
    fct="DiceBCE",
    kwargs = Dict(name="train_loss", use_softmax=USE_SOFTMAX)
)

VAL_LOSS = Dict(
    fct="DiceBCE",
    kwargs = Dict(name="val_loss", use_softmax=USE_SOFTMAX)
)

#---------------------------------------------------------------------------
# metrics configs

TRAIN_METRICS = Dict(
    train_iou=Dict(
        fct="IoU",
        kwargs = Dict(name="train_iou", use_softmax=USE_SOFTMAX)),
    train_dice=Dict(
        fct="Dice",
        kwargs=Dict(name="train_dice", use_softmax=USE_SOFTMAX)),
)

VAL_METRICS = Dict(
    val_iou=Dict(
        fct="IoU",
        kwargs = Dict(name="val_iou", use_softmax=USE_SOFTMAX)),
    val_dice=Dict(
        fct="Dice",
        kwargs=Dict(name="val_dice", use_softmax=USE_SOFTMAX)),
)

#---------------------------------------------------------------------------
# trainers configs

TRAINER = Dict(
    fct="SegTrain",
    kwargs=Dict(),
)

VALIDATER = Dict(
    fct="SegVal",
    kwargs=Dict(),
)

#---------------------------------------------------------------------------
# predictors configs

# PREDICTOR = Dict(
#     fct="SegFast",
#     kwargs=Dict(input_shape=[40,224,224]),
# )
PREDICTOR = Dict(
    fct="SegPatch",
    kwargs=Dict(patch_size=PATCH_SIZE, tta=True, median_spacing=MEDIAN_SPACING, use_softmax=USE_SOFTMAX),
)

############################################################################
# end of config file
# do not write anything in or below this field

CONFIG = Dict(**globals().copy()) # stores all variables in one Dict

to_remove = ['__name__', '__doc__', '__package__',
    '__loader__' , '__spec__', '__annotations__',
    '__builtins__', '__file__', '__cached__', 'Dict']

for k in to_remove: 
    if (k in CONFIG.keys()): del CONFIG[k] 

############################################################################



#---------------------------------------------------------------------------
# A register for all the existing methods
# Aim of this module: 
# - to gather all required imports in a single file
# - to use it in colaboration with a config file
#---------------------------------------------------------------------------

from biom3d.utils import Dict

#---------------------------------------------------------------------------
# dataset register

from biom3d.datasets.semseg_patch_fast import SemSeg3DPatchFast

datasets = Dict(
    SegPatchFast    =Dict(fct=SemSeg3DPatchFast, kwargs=Dict()),
)

#---------------------------------------------------------------------------
# model register

from biom3d.models.unet3d_vgg_deep import UNet
from biom3d.models.encoder_vgg import VGGEncoder, EncoderBlock

models = Dict(
    UNet3DVGGDeep   =Dict(fct=UNet, kwargs=Dict()),
    VGG3D           =Dict(fct=VGGEncoder, kwargs=Dict(block=EncoderBlock, use_head=True)),
)

#---------------------------------------------------------------------------
# metric register

import biom3d.metrics as mt

metrics = Dict(
    Dice        =Dict(fct=mt.Dice, kwargs=Dict()),
    DiceBCE     =Dict(fct=mt.DiceBCE, kwargs=Dict()),
    IoU         =Dict(fct=mt.IoU, kwargs=Dict()),
    MSE         =Dict(fct=mt.MSE, kwargs=Dict()),
    CE          =Dict(fct=mt.CrossEntropy, kwargs=Dict()),
    DeepMSE     =Dict(fct=mt.DeepMetric, kwargs=Dict(metric=mt.MSE)),
    DeepDiceBCE =Dict(fct=mt.DeepMetric, kwargs=Dict(metric=mt.DiceBCE)),
    Triplet     =Dict(fct=mt.Triplet, kwargs=Dict()),
    TripletSeg  =Dict(fct=mt.TripletDiceBCE, kwargs=Dict()),
    ArcFace     =Dict(fct=mt.ArcFace, kwargs=Dict()),
)

#---------------------------------------------------------------------------
# trainer register

from biom3d.trainers import (
    seg_train, 
    seg_validate,
    seg_patch_validate,
    seg_patch_train,
)

trainers = Dict(
    SegTrain        =Dict(fct=seg_train, kwargs=Dict()),
    SegVal          =Dict(fct=seg_validate, kwargs=Dict()),
    SegPatchTrain   =Dict(fct=seg_patch_train, kwargs=Dict()),
    SegPatchVal     =Dict(fct=seg_patch_validate, kwargs=Dict()),
)

#---------------------------------------------------------------------------
# predictor register

from biom3d.predictors import (
    seg_predict,
    seg_predict_patch,
)

predictors = Dict(
    Seg = Dict(fct=seg_predict, kwargs=Dict()),
    SegPatch = Dict(fct=seg_predict_patch, kwargs=Dict()),
)

#---------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# a set of utility functions 
# content:
#  - base class for config file 
#  - read folds from a csv file
#  - create logs and models directories
#  - tif metadata reader and writer
# ----------------------------------------------------------------------------

import numpy as np
from datetime import datetime
from time import time 
import os 
import tifffile as tiff
import matplotlib.pyplot as plt
import yaml # pip install pyyaml
from skimage import io
import SimpleITK as sitk
import torchio as tio

# ----------------------------------------------------------------------------
# read folds from a csv file

def get_train_test_df(df, verbose=True):
    """
    Return the train set and the test set
    """
    train_set = np.array(df[df['hold_out']==0].iloc[:,0])
    test_set = np.array(df[df['hold_out']==1].iloc[:,0])
    return train_set, test_set

def get_folds_df(df, verbose=True):
    """
    Return of folds in a list of list
    """
    folds = []
    if df.empty:
        print("[Warning] one of the data DataFrame is empty!")
        return []
    nbof_folds = df['fold'].max()+1
    if verbose:
        print("Number of folds in df: {}".format(nbof_folds))
    
    size_folds = []
    for i in range(nbof_folds):
        folds += [list(df[df['fold']==i].iloc[:,0])]
        size_folds += [len(folds[-1])]
    if verbose:
        print("Size of folds: {}".format(size_folds))
    return folds

def get_folds_train_test_df(df, verbose=True, merge_test=True):
    """
    Return folds from the train set and the test set in a list of list.
    Output: (train_folds, test_folds)
    If merge_test==True then the test folds are merged in one list.
    """
    if verbose:
        print("Training set:")
    train_folds = get_folds_df(df[df['hold_out']==0], verbose)
    
    if verbose:
        print("Testing set:")
    test_folds = get_folds_df(df[df['hold_out']==1], verbose)
    
    if merge_test:
        test_folds_merged = []
        for i in range(len(test_folds)):
            test_folds_merged += test_folds[i]
        test_folds = test_folds_merged
    return train_folds, test_folds

def get_splits_train_val_test(df):
    """
    the splits contains [100%,50%,25%,10%,5%,2%,the rest] of the dataset
    return the train set as a list of list,
    the val and test set as lists
    """
    nbof_splits = df['split'].max()+1
    valset = list(df[(df['split']==-1)*(df['fold']==0)*(df['hold_out']==0)]['filename'])
    testset = list(df[(df['hold_out']==1)]['filename'])
    train_splits = []
    for i in range(nbof_splits):
        train_splits += [list(df[(df['split']==i)*(df['fold']!=0)*(df['hold_out']==0)].iloc[:,0])]
    # adds the whole dataset in the begging of the train_splits list
    train_splits = [list(df[(df['fold']!=0)*(df['hold_out']==0)].iloc[:,0])] + train_splits
    return train_splits, valset, testset

def get_splits_train_val_test_overlapping(df):
    """
    CAREFUL: works only if the splits contains [1/(2**0), 1/(2**1), ..., 1/(2**n), 1/(2**n)] of the training dataset 
    the splits contains of the dataset.
    "overlapping" indicates that every smaller set is contained into all bigger sets.
    return the train set as a list of list,
    the val and test set as lists
    """
    nbof_splits = df['split'].max()+1
    valset = list(df[(df['split']==-1)*(df['fold']==0)*(df['hold_out']==0)]['filename'])
    testset = list(df[(df['hold_out']==1)]['filename'])
    train_splits = []
    for i in range(nbof_splits):
        train_splits += [list(df[(df['split']>=i)*(df['fold']!=0)*(df['hold_out']==0)].iloc[:,0])]
        
    # adds the last set 
    train_splits += [list(df[(df['split']==(nbof_splits-1))*(df['fold']!=0)*(df['hold_out']==0)].iloc[:,0])]
    return train_splits, valset, testset

# ----------------------------------------------------------------------------
# create logs and models directories

def create_save_dirs(log_dir, desc, dir_names=['model', 'logs', 'images'], return_base_dir=False):
    """
    Creates saving folders. 

    Arguments:
        dir_names: a list of name of the desired folders.
                   e.g.: ['images','cpkt','summary']
    
    Returns:
        list_dirs: a list of path of the corresponding folders.
    """
    list_dirs = []
    current_time = datetime.now().strftime("%Y%m%d-%H%M%S")
    base_dir = current_time + '-' + desc
    base_dir = os.path.join(log_dir, base_dir)
    for name in dir_names:
        list_dirs += [os.path.join(base_dir, name)]
        if not os.path.exists(list_dirs[-1]):
            os.makedirs(list_dirs[-1])
    if return_base_dir:
        return [base_dir] + list_dirs
    else:
        return list_dirs

# ----------------------------------------------------------------------------
# image readers

class skimage_imread:
    """
    image reader for .tif files
    """
    def __call__(self, img_path):
        return io.imread(img_path)

def sitk_imread(img_path):
    """
    image reader for nii.gz files
    """
    img = sitk.ReadImage(img_path)
    img_np = sitk.GetArrayFromImage(img)
    return img_np, np.array(img.GetSpacing())

def adaptive_imread(img_path):
    """
    use skimage imread or sitk imread depending on the file extension:
    .tif --> skimage.io.imread
    .nii.gz --> SimpleITK.imread
    """
    extension = img_path[img_path.rfind('.'):]
    if extension == ".tif":
        return io.imread(img_path)
    else:
        return sitk_imread(img_path)

# ----------------------------------------------------------------------------
# tif metadata reader and writer

def tif_read_meta(tif_path, display=False):
    """
    read the metadata of a tif file and stores them in a python dict.
    if there is a 'ImageDescription' tag, it transforms it as a dictionary
    """
    meta = {}
    with tiff.TiffFile(tif_path) as tif:
        for page in tif.pages:
            for tag in page.tags:
                tag_name, tag_value = tag.name, tag.value
                if display: print(tag.name, tag.code, tag.dtype, tag.count, tag.value)

                # below; fix storage problem for ImageDescription tag
                if tag_name == 'ImageDescription': 
                    list_desc = tag_value.split('\n')
                    dict_desc = {}
                    for idx, elm in enumerate(list_desc):
                        split = elm.split('=')
                        dict_desc[split[0]] = split[1]
                    meta[tag_name] = dict_desc
                else:
                    meta[tag_name] = tag_value
            break # just check the first image
    return meta

def tif_write_meta(data,meta,out_path):
    """
    write data and metadata in 'out_path'
    """
    out_meta = {
        'spacing':float(meta['ImageDescription']['spacing']),
        'unit':meta['ImageDescription']['unit'],
        'axes':'ZYX',
    }
    
    extratags = []
    
    tiff.imwrite(
        out_path,
        data=data,
        resolution=(meta['XResolution'],meta['YResolution']),
        metadata=out_meta,
        extratags=extratags,
        imagej=True,
    )

def tif_copy_meta(in_path1, in_path2, out_path):
    """
    store (metadata of in_path1 + data of in_path2) in out_path
    """
    in_meta = tif_read_meta(in_path1)
    data = tiff.imread(in_path2)
    tif_write_meta(data, in_meta, out_path)

def tif_get_spacing(path):
    """
    get the image spacing stored in the metadata file.
    """
    img_meta = tif_read_meta(path)

    xres = (img_meta["XResolution"][1]/img_meta["XResolution"][0])*1e-6
    yres = (img_meta["YResolution"][1]/img_meta["YResolution"][0])*1e-6
    zres = float(img_meta["ImageDescription"]["spacing"])*1e-6
    max_dim = min([xres,yres,zres])
    xres = max_dim / xres
    yres = max_dim / yres
    zres = max_dim / zres
    return (xres, yres, zres)

# ----------------------------------------------------------------------------
# 3d viewer

def display_voxels(image, xlim, ylim, zlim, save=False):
    """
    plot using matplotlib a 3d volume from a 3d image
    """
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.voxels(image)
    
    ax.set_xlim(xlim[0], xlim[1])  
    ax.set_ylim(ylim[0], ylim[1])  
    ax.set_zlim(zlim[0], zlim[1])
    
    plt.tight_layout()
    plt.savefig('voxel.png') if save else plt.show() 

def display_mesh(mesh, xlim, ylim, zlim, save=False):
    """
    plot using matplotlib a 3d volume from a 3d mesh
    """
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.add_collection3d(mesh)
    
    ax.set_xlim(xlim[0], xlim[1])  
    ax.set_ylim(ylim[0], ylim[1])  
    ax.set_zlim(zlim[0], zlim[1])
    
    plt.tight_layout()
    plt.savefig('mesh.png') if save else plt.show() 

def napari_viewer(img, pred):
    import napari
    viewer = napari.view_image(img, name='original')
    viewer.add_image(pred, name='pred')
    viewer.layers['pred'].opacity=0.5
    viewer.layers['pred'].colormap='red'
    napari.run()

# ----------------------------------------------------------------------------
# os utils

def abs_path(root, listdir_):
    listdir = listdir_.copy()
    for i in range(len(listdir)):
        listdir[i] = os.path.join(root, listdir[i])
    return listdir

def abs_listdir(path):
    return abs_path(path, os.listdir(path))

# ----------------------------------------------------------------------------
# preprocess utils
# from the median image shape predict the size of the patch, the pool, the batch 

def single_patch_pool(dim, size_limit=7):
    """
    divide by two the dim number until obtaining a number lower than 7
    then np.ceil this number
    then multiply multiple times by two this number to obtain the patch size
    """
    pool = 0
    while dim > size_limit:
        dim /= 2
        pool += 1
    patch = np.round(dim)
    patch = patch.astype(int)*(2**pool)
    return patch, pool

def find_patch_pool_batch(dims, max_dims=(128,128,128), max_pool=5, epsilon=1e-3):
    """
    take the median size as input, determine the patch size and the number of pool
    with "single_patch_pool" function for each dimension and 
    assert that the final dimension size is smaller than max_dims.prod().
    """
    # transform tuples into arrays
    dims = np.array(dims)
    max_dims = np.array(max_dims)
    
    # divides by a 1+epsilon until reaching a sufficiently small resolution
    while dims.prod() > max_dims.prod():
        dims = dims / (1+epsilon)
    dims = dims.astype(int)
    
    # compute patch and pool for all dims
    patch_pool = np.array([single_patch_pool(m) for m in dims])
    patch = patch_pool[:,0]
    pool = patch_pool[:,1]
    
    # assert the final size is smaller than max_dims
    while patch.prod()>max_dims.prod():
        patch = patch - np.array([32,32,32])*(patch>max_dims) # removing multiples of 32
    pool = np.where(pool > max_pool, max_pool, pool)
    
    # batch_size
    batch = 2
    while batch*patch.prod() <= 2*max_dims.prod():
        batch += 1
    if batch*patch.prod() > 2*max_dims.prod():
        batch -= 1
    return patch, pool, batch

def convert_num_pools(num_pools):
    """
    Set adaptive number of pools
        for example: convert [3,5,5] into [[1 1 1],[1 2 2],[2 2 2],[2 2 2],[2 2 2],[1 2 2]]
    """
    max_pool = max(num_pools)
    strides = []
    for i in range(len(num_pools)):
        st = np.ones(max_pool)
        num_zeros = max_pool-num_pools[i]
        for j in range(num_zeros):
            st[j]=0
        st=np.roll(st,-num_zeros//2)
        strides += [st]
    strides = np.array(strides).astype(int).T+1
    # kernels = (strides*3//2).tolist()
    strides = strides.tolist()
    return strides

# ----------------------------------------------------------------------------
# data augmentation utils

def centered_pad(img, final_size, msk=None):
    """
    centered pad an img and msk to fit the final_size
    """
    final_size = np.array(final_size)
    img_shape = np.array(img.shape[1:])
    
    start = (final_size-np.array(img_shape))//2
    end = final_size-(img_shape+start)
    end = end * (end > 0)
    
    pad = np.append([[0,0]], np.stack((start,end),axis=1), axis=0)
    pad_img = np.pad(img, pad, 'constant', constant_values=0)
    if msk is not None: pad_msk = np.pad(msk, pad, 'constant', constant_values=0)
    
    if msk is not None:
        return pad_img, pad_msk
    else: 
        return pad_img

class RandomCropResize:
    """
    SmartPatch
    Randomly crop and resize the images to a certain crop_shape.
    The global_crop_resize method performs a random crop and resize.
    The local_crop_resize method performs a random crop and resize making sure that the crop 
    is overlapping (to a certain extent, defined by the min_overlap parameter) with the global
    crop previously performed. 
    """
    def __init__(
        self,
        local_crop_shape,
        global_crop_shape,
        min_overlap,
        global_crop_scale=1.0,
        global_crop_min_shape_scale=1.0,
        ):
        """
        Parameters
        ----------
        global_crop_shape : list or tuple of size == 3
            Minimal crop size
        global_crop_scale : float, default=1.0
            Value between 0 and 1. Factor multiplying (img_shape - global_crop_min_shape) and added to the global_crop_min_shape. A value of 1 means that the maximum shape of the global crop will be the image shape. A value of 0 means that the maximum value will be the global_crop_min_shape. 
        global_crop_min_shape_factor : float, default=1.0
            (DEPRECATED?) Factor multiplying the minimal global_crop_shape, 1.0 is a good default
        
        """
        
        self.local_crop_shape = np.array(local_crop_shape)
        self.global_crop_shape = np.array(global_crop_shape)
        self.global_crop_scale = np.array(global_crop_scale)
        self.global_crop_min_shape_scale = np.array(global_crop_min_shape_scale)
        self.alpha = 1  - min_overlap
        
        # internal arguments
        self.global_crop_center = None
        
    def global_crop_resize(self, img, msk=None):
        img_shape = np.array(img.shape)[1:]
        
        # determine crop shape
        min_crop_shape = np.round(self.global_crop_shape * self.global_crop_min_shape_scale).astype(int)
        min_crop_shape = np.minimum(min_crop_shape, img_shape)
        crop_shape = np.random.randint(min_crop_shape, (img_shape-min_crop_shape)*self.global_crop_scale+min_crop_shape+1)
        
        # determine crop coordinates
        rand_start = np.random.randint(0, np.maximum(1,img_shape-crop_shape))
        rand_end = crop_shape+rand_start
        
        self.global_crop_center = (rand_end-rand_start)//2 + rand_start
        
        # crop
        crop_img = img[:,
                        rand_start[0]:rand_end[0], 
                        rand_start[1]:rand_end[1], 
                        rand_start[2]:rand_end[2]]
        
        if msk is not None:
            crop_msk = msk[:,
                            rand_start[0]:rand_end[0], 
                            rand_start[1]:rand_end[1], 
                            rand_start[2]:rand_end[2]]
    
        # temp: resize must be done!
        if not np.array_equal(crop_img.shape[1:], self.global_crop_shape):
            if msk is not None:
                sub = tio.Subject(img=tio.ScalarImage(tensor=crop_img), msk=tio.LabelMap(tensor=crop_msk))
                sub = tio.Resize(self.global_crop_shape)(sub)
                crop_img, crop_msk = sub.img.tensor, sub.msk.tensor
            else:
                crop_img = tio.Resize(self.global_crop_shape)(crop_img)
        
        # returns
        if msk is not None:
            return crop_img, crop_msk
        else:
            return crop_img

    def local_crop_pad(self, img, msk=None):
        """
        global_crop_resize must be called at least once before calling local_crop_pad
        """
        assert self.global_crop_center is not None, "Error! self.global_crop_resize must be called once before self.local_crop_pad."
        
        img_shape = np.array(img.shape)[1:]
        crop_shape = self.local_crop_shape
        
        # determine crop coordinates
        # we make sure that the crop shape overlap with the global crop shape by at least min_overlap
        centers_max_dist = np.round(crop_shape * self.alpha).astype(np.uint8) + (self.global_crop_shape-crop_shape)//2
        local_center_low = np.maximum(crop_shape//2, self.global_crop_center-centers_max_dist)
        local_center_high = np.minimum(img_shape - crop_shape//2, self.global_crop_center+centers_max_dist)
        local_center_high = np.maximum(local_center_high, local_center_low+1)

        local_crop_center = np.random.randint(low=local_center_low, high=local_center_high)
        
        # local
        start = local_crop_center - (self.local_crop_shape//2)
        start = np.maximum(0,start)
        end = start + self.local_crop_shape
        
        crop_img = img[:,
                    start[0]:end[0], 
                    start[1]:end[1], 
                    start[2]:end[2]]
        
        if msk is not None:
            crop_msk = msk[:,
                        start[0]:end[0], 
                        start[1]:end[1], 
                        start[2]:end[2]]
        
        # pad if needed
        if not np.array_equal(crop_img.shape[1:], self.local_crop_shape):
            if msk is not None:
                crop_img, crop_msk = centered_pad(img=crop_img, final_size=self.local_crop_shape, msk=crop_msk)
            else:
                crop_img = centered_pad(img=crop_img, final_size=self.local_crop_shape)
        
        # returns
        if msk is not None:
            return crop_img, crop_msk
        else:
            return crop_img

    def local_crop_resize(self, img, msk=None):
        """
        global_crop_resize must be called at least once before calling local_crop_resize
        """
        assert self.global_crop_center is not None, "Error! self.global_crop_resize must be called once before self.local_crop_resize."

        img_shape = np.array(img.shape)[1:]

        # determine crop shape
        crop_shape = np.random.randint(self.local_crop_scale[0] * img_shape, self.local_crop_scale[1] * img_shape+1)
        
        # determine crop coordinates
        # we make sure that the crop shape overlap with the global crop shape by at least min_overlap
        centers_max_dist = np.round(crop_shape * self.alpha).astype(np.uint8) + (self.global_crop_shape-crop_shape)//2
        local_center_low = np.maximum(crop_shape//2, self.global_crop_center-centers_max_dist)
        local_center_high = np.minimum(img_shape - crop_shape//2, self.global_crop_center+centers_max_dist)
        local_center_high = np.maximum(local_center_high, local_center_low+1)

        local_crop_center = np.random.randint(low=local_center_low, high=local_center_high)
        
        start = local_crop_center - (self.local_crop_shape//2)
        start = np.maximum(0,start)
        end = start + self.local_crop_shape
        
        crop_img = img[:,
                    start[0]:end[0], 
                    start[1]:end[1], 
                    start[2]:end[2]]
        
        if msk is not None:
            crop_msk = msk[:,
                        start[0]:end[0], 
                        start[1]:end[1], 
                        start[2]:end[2]]
        
        # resize if needed
        if not np.array_equal(crop_img.shape[1:], self.local_crop_shape):
            if msk is not None:
                sub = tio.Subject(img=tio.ScalarImage(tensor=crop_img), msk=tio.LabelMap(tensor=crop_msk))
                sub = tio.Resize(self.global_crop_shape)(sub)
                crop_img, crop_msk = sub.img.tensor, sub.msk.tensor
            else:
                crop_img = tio.Resize(self.global_crop_shape)(crop_img)
        
        # returns
        if msk is not None:
            return crop_img, crop_msk
        else:
            return crop_img

# ----------------------------------------------------------------------------
# config utils
# Convenience class that behaves exactly like dict(), but allows accessing
# the keys and values using the attribute syntax, i.e., "mydict.key = value".
# Author: Terro Keras (progressive_growing_of_gans)

class Dict(dict):
    def __init__(self, *args, **kwargs): super().__init__(*args, **kwargs)
    def __getattr__(self, name): return self[name]
    def __setattr__(self, name, value): self[name] = value
    def __delattr__(self, name): del self[name]

def Dict_to_dict(cfg):
    """
    transform a Dict into a dict
    """
    ty = type(cfg)
    cfg = dict(cfg)
    for k,i in cfg.items():
        if type(i)==ty:
            cfg[k] = Dict_to_dict(cfg[k])
    return cfg

def dict_to_Dict(cfg):
    """
    transform a Dict into a dict
    """
    ty = type(cfg)
    cfg = Dict(cfg)
    for k,i in cfg.items():
        if type(i)==ty:
            cfg[k] = dict_to_Dict(cfg[k])
    return cfg

def save_config(path, cfg):
    """
    save a configuration in a yaml file.
    path must thus contains a yaml extension.
    example: path='logs/test.yaml'
    """
    cfg = Dict_to_dict(cfg)
    with open(path, "w") as f:
        yaml.dump(cfg, f, sort_keys=False)
    
def load_config(path):
    """
    load a yaml stored with the self.save method.
    """
    return dict_to_Dict(yaml.load(open(path),Loader=yaml.FullLoader))

def nested_dict_pairs_iterator(dic):
    ''' This function accepts a nested dictionary as argument
        and iterate over all values of nested dictionaries
        stolen from: https://thispointer.com/python-how-to-iterate-over-nested-dictionary-dict-of-dicts/ 
    '''
    # Iterate over all key-value pairs of dict argument
    for key, value in dic.items():
        # Check if value is of dict type
        if isinstance(value, dict) or isinstance(value, Dict):
            # If value is dict then iterate over all its values
            for pair in  nested_dict_pairs_iterator(value):
                yield [key, *pair]
        else:
            # If value is not dict type then yield the value
            yield [key, value]

def nested_dict_change_value(dic, key, value):
    """
    Change all value with a given key from a nested dictionary.
    """
    # Loop through all key-value pairs of a nested dictionary and change the value 
    for pairs in nested_dict_pairs_iterator(dic):
        if key in pairs:
            save = dic[pairs[0]]; i=1
            while i < len(pairs) and pairs[i]!=key:
                save = save[pairs[i]]; i+=1
            save[key] = value
    return dic

# ----------------------------------------------------------------------------
# postprocessing utils

from skimage import measure
def dist_vec(v1,v2):
    """
    euclidean distance between two vectors (np.array)
    """
    v = v2-v1
    return np.sqrt(np.sum(v*v))

def center(labels, idx):
    """
    return the barycenter of the pixels of label = idx
    """
    dim = labels.shape
    
    # matrix of coordinate with the same size as labels
    x, y, z = np.meshgrid(np.arange(dim[1]),np.arange(dim[0]), np.arange(dim[2]))
    out = np.stack((y,x,z))
    out = np.transpose(out, axes=(1,2,3,0))
    
    # extract the barycenter
    return np.mean(out[labels==idx], axis=0)

def closest(labels, num):
    """
    return the index of the object the closest to the center of the image.
    num: number of label in the image (background does not count)
    """
    labels_center = np.array(labels.shape)/2
    centers = [center(labels,idx+1) for idx in range(num)]
    dist = [dist_vec(labels_center,c) for c in centers]
    return np.argmin(dist)+1

def keep_center_only(msk):
    """
    return mask (msk) with only the connected component that is the closest 
    to the center of the image.
    """
    labels, num = measure.label(msk, background=0, return_num=True)
    close_idx = closest(labels,num)
    return (labels==close_idx).astype(msk.dtype)*255

def volumes(labels):
    """
    returns the volumes of all the labels in the image
    """
    return [((labels==idx).astype(int)).sum() for idx in np.unique(labels)]

def keep_biggest_volume_centered(msk):
    """
    return mask (msk) with only the connected component that is the closest 
    to the center of the image if its volumes is not too small ohterwise returns
    the biggest object (different from the background).
    (too small meaning that its volumes shouldn't smaller than half of the biggest one)
    the final mask intensities are either 0 or msk.max()
    """
    labels, num = measure.label(msk, background=0, return_num=True)
    close_idx = closest(labels,num)
    vol = volumes(labels)
    relative_vol = [vol[close_idx]/vol[idx] for idx in range(1,len(vol))]
    min_rel_vol = np.min(relative_vol)
    if min_rel_vol < 0.5:
        close_idx = np.argmin(relative_vol)+1
    return (labels==close_idx).astype(msk.dtype)*msk.max()

# ----------------------------------------------------------------------------
# test utils

def adaptive_imread(img_path):
    """
    use skimage imread or sitk imread depending on the file extension:
    .tif --> skimage.io.imread
    .nii.gz --> SimpleITK.imread
    """
    extension = img_path[img_path.rfind('.'):]
    if extension == ".tif":
        return io.imread(img_path), []
    else:
        return sitk_imread(img_path)

def one_hot(values, num_classes=None):
    """
    One hot encoding
    """
    if num_classes==None:
        n_values = np.max(values) + 1
    else:
        n_values = num_classes
    out = np.eye(n_values)[values]
    out = np.moveaxis(out, -1, 0)
    return out

# metric definition
def iou(inputs, targets, smooth=1):
    inter = (inputs & targets).sum()
    union = (inputs | targets).sum()
    return (inter+smooth)/(union+smooth)

def dice(inputs, targets, smooth=1):   
    inter = (inputs & targets).sum()                           
    dice = (2.*inter + smooth)/(inputs.sum() + targets.sum() + smooth)  
    return dice

def versus_one(fct, in_path, tg_path, num_classes, single_class=None):
    """
    comparison function between in_path image and tg_path and using the criterion defined by fct
    """
    img1,_ = adaptive_imread(in_path)
    print("input path",in_path)
    if len(img1.shape)==3:
        img1 = one_hot(img1, num_classes)[1:,...]
    if single_class is not None:
        img1 = img1[single_class,...]
    img1 = (img1 > 0).astype(int)
    
    img2,_ = adaptive_imread(tg_path)
    print("target path",tg_path)
    if len(img2.shape)==3:
        img2 = one_hot(img2, num_classes)[1:,...]
    if single_class is not None:
        img2 = img2[single_class,...]
    img2 = (img2 > 0).astype(int)
    
    # remove background if needed
    if img1.shape[0]==(img2.shape[0]+1):
        img1 = img1[1:]
    if img2.shape[0]==(img1.shape[0]+1):
        img2 = img2[1:]
    
    if sum(img1.shape)!=sum(img2.shape):
        print("bug:sum(img1.shape)!=sum(img2.shape):")
        print("img1.shape", img1.shape)
        print("img2.shape", img2.shape)
        return
    out = fct(img1, img2)
    return out

# def versus(fct, list_abs1,list_abs2, num_classes=None, single_class=None):
#     """
#     Computes the function 'fct' over the whole list_abs1 and list_abs2
#     args:
#         -num_classes: [int] number of classes in the images (number of channels)
#         -single_class: [int] if not None, used to select one of the predicted classes.
#         must be smaller than the number of classes.
#     """
#     mean = 0
#     count = 0
#     for i in range(len(list_abs1)):
#         img1 = adaptive_imread(list_abs1[i])
#         img1 = one_hot(img1, num_classes)[...,1:]
#         if single_class is not None:
#             img1 = img1[...,single_class]
# #         img1 = keep_center_only(img1)
#         img1 = (img1 > 0).astype(int)
    
#         img2 = adaptive_imread(list_abs2[i])
#         img2 = one_hot(img2, num_classes)[...,1:]
#         if single_class is not None:
#             img2 = img2[...,single_class]
# #         img2 = keep_center_only(img2)
#         img2 = (img2 > 0).astype(int)

#         if sum(img1.shape)!=sum(img2.shape):
#             print("bug:sum(img1.shape)!=sum(img2.shape):", list_abs1[i], list_abs2[i])
#             print("img1.shape", img1.shape)
#             print("img2.shape", img2.shape)
#             pass
#         else:
#             out = fct(img1,img2)
# #             if out!=1:
# #                 print(out)
# #                 print(i)
# #                 print(list_abs1[i])
# #                 print(list_abs2[i])
#             mean += out
            
#             count += 1
#     return mean / count

# def versus_one(fct, in_path, tg_path, num_classes, single_class=None):
#     img1 = adaptive_imread(in_path)
#     img1 = one_hot(img1, num_classes)[...,1:]
#     if single_class is not None:
#         img1 = img1[...,single_class]
#     img1 = (img1 > 0).astype(int)
    
#     img2 = adaptive_imread(tg_path)
#     img2 = one_hot(img2, num_classes)[...,1:]
#     if single_class is not None:
#         img2 = img2[...,single_class]
#     img2 = (img2 > 0).astype(int)
    

#     print("img1.shape", img1.shape)
#     print("img2.shape", img2.shape)
#     if sum(img1.shape)!=sum(img2.shape):
#         print("bug:sum(img1.shape)!=sum(img2.shape):", list_abs[i])
#         print("img1.shape", img1.shape)
#         print("img2.shape", img2.shape)
#         return
#     out = fct(img1, img2)
#     return out

# def versus_all_sophie(versus, fct, list_abs):
#     """
#     use the "versus" function to apply the "fct" function over the 
#     different elements in list "list_abs".
#     """
#     print("sophie vs mine", versus(fct, list_abs[0], list_abs[1]))
#     print("sophie vs gift", versus(fct, list_abs[0], list_abs[2]))
#     print("sophie vs otsu", versus(fct, list_abs[0], list_abs[3]))

#     print("mine vs gift", versus(fct, list_abs[1], list_abs[2]))
#     print("mine vs otsu", versus(fct, list_abs[1], list_abs[3]))

#     print("otsu vs gift", versus(fct, list_abs[2], list_abs[3]))

# def versus_all(versus, fct, list_abs, list_names=None):
#     """
#     use the "versus" function to apply the "fct" function over the 
#     different elements in list "list_abs".
#     """
#     if list_names is None:
#         list_names = ['lab_{}'.format(i) for i in range(len(list_abs)-1)]
#     for i in range(len(list_abs)-1):
#         for j in range(i):
#             print("{} vs {}:".format(list_names[i], list_names[j]), versus(fct, list_abs[i], list_abs[j]))

# def one_versus_all(versus, fct, list_abs, list_names=None, num_classes=None, single_class=None):
#     """
#     use the "versus" function to apply the "fct" function over the 
#     different elements in list "list_abs" .
#     one vs all is the comparison between the label and the rest of the images
#     """
# #     if list_names is None:
# #         list_names = ['lab_{}'.format(i) for i in range(len(list_abs)-1)]
#     for i in range(1,len(list_abs)-1):
#         print("{} vs {}:".format(list_names[0], list_names[i]), versus(fct, list_abs[0], list_abs[i], num_classes, single_class))

# ----------------------------------------------------------------------------
# time utils

class Time:
    def __init__(self, name=None):
        self.name=name
        self.reset()
    
    def reset(self):
        print("Count has been reset!")
        self.start_time = time()
        self.count = 0
    
    def get(self):
        self.count += 1
        return time()-self.start_time
    
    def __str__(self):
        self.count += 1
        out = time() - self.start_time
        self.start_time=time()
        return "[DEBUG] name: {}, count: {}, time: {} seconds".format(self.name, self.count, out)

# ----------------------------------------------------------------------------



#---------------------------------------------------------------------------
# Main code: run predictions
#---------------------------------------------------------------------------

import os
import argparse
import pathlib
import numpy as np
# from telegram_send import send
# from magicgui import magicgui

from biom3d.builder import Builder
from biom3d.utils import abs_listdir, versus_one, dice

#---------------------------------------------------------------------------
# prediction base fonction

def pred(bui_dir, dir_in, dir_out):
    bui_dir=str(bui_dir)
    dir_in=str(dir_in)
    dir_out=str(dir_out)
    LOG_PATH = bui_dir

    builder = Builder(config=None,path=LOG_PATH, training=False)
    builder.run_prediction_folder(dir_in=dir_in, dir_out=dir_out, return_logit=False)

def pred_multiple(bui_dir, dir_in, dir_out):
    """
    predict a folder of folders
    """
    list_dir_in = [dir_in + "/" + e for e in os.listdir(dir_in)]
    list_dir_out = [dir_out + "/" + e for e in os.listdir(dir_in)]
    LOG_PATH = bui_dir

    for i in range(len(list_dir_in)):
        dir_in = list_dir_in[i]
        dir_out = list_dir_out[i]

        builder = Builder(config=None,path=LOG_PATH, training=False)
        builder.run_prediction_folder(dir_in=dir_in, dir_out=dir_out, return_logit=False)

#---------------------------------------------------------------------------
# main unet segmentation

# import configs.config_unet as config_unet

# @magicgui(call_button="predict")
def pred_seg(bui_dir=pathlib.Path.home(), dir_in=pathlib.Path.home(), dir_out=pathlib.Path.home()):
    pred(bui_dir, dir_in, dir_out)

def pred_seg_eval(bui_dir=pathlib.Path.home(), dir_in=pathlib.Path.home(), dir_out=pathlib.Path.home(), dir_lab=None, eval_only=False):
    print("Start inference")
    builder_pred = Builder(
        config=None,
        path=bui_dir, 
        training=False)

    dir_out = os.path.join(dir_out,os.path.split(bui_dir)[-1]) # name the prediction folder with the model folder name
    if not eval_only:
        builder_pred.run_prediction_folder(dir_in=dir_in, dir_out=dir_out, return_logit=False) # run the predictions
    print("Inference done!")


    if dir_lab is not None:
        # eval
        print("Start evaluation")
        paths_lab = [dir_lab, dir_out]
        list_abs = [sorted(abs_listdir(p)) for p in paths_lab]
        assert sum([len(t) for t in list_abs])%len(list_abs)==0, "[Error] Not the same number of labels and predictions! {}".format([len(t) for t in list_abs])

        results = []
        for idx in range(len(list_abs[0])):
            print("Metric computation for:", list_abs[1][idx])
            results += [versus_one(
                fct=dice, 
                in_path=list_abs[1][idx], 
                tg_path=list_abs[0][idx], 
                # num_classes=2, 
                # single_class=-1,
                num_classes=builder_pred.config.NUM_CLASSES if builder_pred.config.USE_SOFTMAX else (builder_pred.config.NUM_CLASSES+1), 
                single_class=None,
                )]
            print("Metric result:", print(results[-1]))
        print("Evaluation done! Average result:", np.mean(results))
        # send(messages=["Evaluation done of model {}! Average result: {}".format(dir_out, np.mean(results))])

#---------------------------------------------------------------------------

if __name__=='__main__':

    # methods names 
    valid_names = {
        "seg": pred_seg,
        "seg_eval": pred_seg_eval,
        "seg_multiple": pred_multiple,
        # "seg_patch": pred_seg_patch,
        # "seg_patch_multi": pred_seg_patch_multi,
        # "single": pred_single,
        # "triplet": main_triplet,
        # "arcface": main_arcface,
        # "unet_triplet": main_unet_triplet,
        # "cotrain": main_cotrain,
        # "cotrain_and_single": main_cotrain_and_single
    }

    # parser
    parser = argparse.ArgumentParser(description="Main training file.")
    parser.add_argument("-n", "--name", type=str, default="seg",
        help="Name of the tested method. Valid names: {}".format(valid_names.keys()))
    parser.add_argument("-b", "--bui_dir", type=str,
        help="Path of the builder directory")
    parser.add_argument("-i", "--dir_in", type=str,
        help="Path to the input image directory")
    parser.add_argument("-o", "--dir_out", type=str,
        help="Path to the output prediction directory")
    parser.add_argument("-a", "--dir_lab", type=str, default=None,
        help="Path to the input image directory") 
    parser.add_argument("-e", "--eval_only", default=False,  action='store_true', dest='eval_only',
        help="Do only the evaluation and skip the prediction (predictions must have been done already.)") 
    args = parser.parse_args()

    # run the method
    # valid_names[args.name](
    #     bui_dir=args.bui_dir,
    #     dir_in=args.dir_in,
    #     dir_out=args.dir_out,
    # )
    if args.bui_dir is None:
        valid_names[args.name].show(run=True)
    else:
        if args.dir_lab is not None:
            valid_names[args.name](args.bui_dir, args.dir_in, args.dir_out, args.dir_lab, args.eval_only)
        else:
            valid_names[args.name](args.bui_dir, args.dir_in, args.dir_out)

#---------------------------------------------------------------------------


#---------------------------------------------------------------------------
# Metrics/losses for segmentation
#---------------------------------------------------------------------------

import torch 
from torch import nn 
import torch.nn.functional as F
import numpy as np

#---------------------------------------------------------------------------
# Metrics base class

class Metric(nn.Module):
    """
    base class for metric to store its value and average and name.
    """
    def __init__(self, name=None):
        super(Metric, self).__init__()
        self.name = name
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
    
    def forward(self, preds, trues):
        pass

    def update(self, n=1):
        with torch.no_grad():
            self.sum += self.val * n
            self.count += n
            self.avg = self.sum / self.count
    
    def str(self):
        return str(self.val)
    
    def __str__(self):
        to_print = self.avg if self.avg!=0 else self.val
        return "{} {:.3f}".format(self.name, to_print)

#---------------------------------------------------------------------------
# Metrics/losses for semantic segmentation

class Dice(Metric):
    def __init__(self, use_softmax=False, name=None):
        super(Dice, self).__init__()
        self.name = name
        self.use_softmax = use_softmax # if use softmax then remove bg

    def forward(self, inputs, targets, smooth=1):
        if self.use_softmax:
            inputs = inputs.softmax(dim=1)[:,1:]
            targets = targets[:,1:]
        else:
            inputs = inputs.sigmoid()

        #flatten label and prediction tensors
        inputs = inputs.reshape(-1)
        targets = targets.reshape(-1)

        intersection = (inputs * targets).sum()                            
        dice = (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)  

        self.val = 1 - dice if self.training else dice
        return self.val 
    
class DiceBCE(Metric):
    """
    num_classes should be > 1 only if softmax use is intended!
    """
    def __init__(self, use_softmax=False, name=None):
        super(DiceBCE, self).__init__()
        self.use_softmax = use_softmax # if use softmax then remove bg for dice computation
        self.name = name
        self.bce = torch.nn.CrossEntropyLoss(reduction='mean')

    def forward(self, inputs, targets, smooth=1):
        #comment out if your model contains a sigmoid or equivalent activation layer
        if self.use_softmax:
            targets_bce = targets.argmax(dim=1).long()
        
            BCE = self.bce(inputs, targets_bce)

            # for dice computation, remove the background and flatten
            inputs = inputs.softmax(dim=1)[:,1:].reshape(-1)
            targets = targets[:,1:].reshape(-1)

        else:
            # keep the background and flatten
            inputs = inputs.reshape(-1)
            targets = targets.reshape(-1)
            BCE = F.binary_cross_entropy_with_logits(inputs, targets, reduction='mean')
            inputs = inputs.sigmoid()


        intersection = (inputs * targets).sum()                            
        dice_loss = 1 - (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)
        Dice_BCE = BCE + dice_loss
        
        self.val = Dice_BCE
        return self.val 
    
class IoU(Metric):
    def __init__(self, use_softmax=False, name=None):
        super(IoU, self).__init__()
        self.use_softmax = use_softmax # if use softmax then remove bg
        self.name = name 

    def forward(self, inputs, targets, smooth=1):
        if self.use_softmax:
            inputs = inputs.softmax(dim=1)[:,1:]
            targets = targets[:,1:]
        else:
            inputs = inputs.sigmoid()

        inputs = inputs.reshape(-1)
        targets = targets.reshape(-1)       
        
        #intersection is equivalent to True Positive count
        #union is the mutually inclusive area of all labels & predictions 
        intersection = (inputs * targets).sum()
        total = (inputs + targets).sum()
        union = total - intersection 
        
        iou = (intersection + smooth)/(union + smooth)
        
        self.val = 1 - iou if self.training else iou
        return self.val 

class MSE(Metric):
    def __init__(self, name=None):
        super(MSE, self).__init__()
        self.name = name

    def forward(self, inputs, targets):

        self.val = torch.nn.functional.mse_loss(inputs, targets, reduction='mean')
        return self.val

class CrossEntropy(Metric):
    def __init__(self, name=None):
        super(CrossEntropy, self).__init__()
        self.name = name
        self.ce = torch.nn.CrossEntropyLoss(reduction='mean')

    def forward(self, inputs, targets):
        self.val = self.ce(inputs, targets)
        return self.val

#---------------------------------------------------------------------------
# Loss for self-supervised pre-training

class Triplet(Metric):
    """
    triplet loss.
    """
    def __init__(self, alpha=0.2, name=None):
        super(Triplet, self).__init__()
        self.alpha = alpha
        self.name = name 

    def forward(self, anchor, positive, negative, smooth=1):
        # normalize
        anc_normed = F.normalize(anchor)
        pos_normed = F.normalize(positive)
        neg_normed = F.normalize(negative)

        dist_pos = torch.sum(torch.square(anc_normed-pos_normed), dim=1)
        dist_neg = torch.sum(torch.square(anc_normed-neg_normed), dim=1)

        self.val = torch.mean(F.relu(dist_pos-dist_neg+self.alpha))
        return self.val

class TripletDiceBCE(Metric):
    """
    this class is just a placeholder for the triplet loss and the DiceBCE loss
    """
    def __init__(self, alpha=0.2, use_softmax=False, name=None):
        super(TripletDiceBCE, self).__init__()

        self.name = name 
        self.triplet_loss = Triplet(alpha=alpha, name='triplet_loss')
        self.dice_loss = DiceBCE(use_softmax=use_softmax, name='dice_loss')

    def update_val(self):
        """
        utils function to update self.val value
        """
        self.val = self.triplet_loss.val + self.dice_loss.val 

#---------------------------------------------------------------------------
# ArcFace losses
# stolen from: https://github.com/ronghuaiyang/arcface-pytorch/blob/master/models/metrics.py 

import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Parameter
import math

class ArcFace(Metric):
    """ 
    from: https://github.com/ydwen/opensphere/blob/main/model/head/arcface.py 
    reference: <Additive Angular Margin Loss for Deep Face Recognition>
    """
    def __init__(self, s=64., m=0.5, name=None):
        super(ArcFace, self).__init__()

        self.s = s
        self.m = m
        self.criterion = torch.nn.CrossEntropyLoss(reduction='mean')

        self.name = name

    def forward(self, x, y):
        with torch.no_grad():
            theta_m = torch.acos(x.clamp(-1+1e-5, 1-1e-5))
            theta_m.scatter_(1, y.view(-1, 1), self.m, reduce='add')
            theta_m.clamp_(1e-5, 3.14159)
            d_theta = torch.cos(theta_m) - x

        logits = self.s * (x + d_theta)
        self.val = self.criterion(logits, y)

        return self.val

class ArcMarginProduct(nn.Module):
    r"""Implement of large margin arc distance: :
        Args:
            in_features: size of each input sample
            out_features: size of each output sample
            s: norm of input feature
            m: margin
            cos(theta + m)
        """
    def __init__(self, in_features, out_features, s=30.0, m=0.50, easy_margin=False):
        super(ArcMarginProduct, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.s = s
        self.m = m
        self.weight = Parameter(torch.FloatTensor(out_features, in_features))
        nn.init.xavier_uniform_(self.weight)

        self.easy_margin = easy_margin
        self.cos_m = math.cos(m)
        self.sin_m = math.sin(m)
        self.th = math.cos(math.pi - m)
        self.mm = math.sin(math.pi - m) * m

    def forward(self, input, label):
        # --------------------------- cos(theta) & phi(theta) ---------------------------
        cosine = F.linear(F.normalize(input), F.normalize(self.weight))
        sine = torch.sqrt((1.0 - torch.pow(cosine, 2)).clamp(0, 1))
        phi = cosine * self.cos_m - sine * self.sin_m
        if self.easy_margin:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where(cosine > self.th, phi, cosine - self.mm)
        # --------------------------- convert label to one-hot ---------------------------
        one_hot = F.one_hot(label, num_classes=self.out_features)
        # -------------torch.where(out_i = {x_i if condition_i else y_i) -------------
        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)  # you can use torch.where if your torch.__version__ is 0.4
        output *= self.s

        return output

# class ArcFace(Metric):
#     """
#     CAREFUL: this module has parameters!!
#     """
#     def __init__(self, in_features, out_features, s=30.0, m=0.50, easy_margin=False, name=None):
#         super(ArcFace, self).__init__()
#         self.name = name
#         self.arc = ArcMarginProduct(in_features=in_features, out_features=out_features, s=s, m=m, easy_margin=easy_margin)
#         self.criterion = torch.nn.CrossEntropyLoss()
        
#     def forward(self, input, label):
#         self.val = self.criterion(self.arc(input, label), label)
#         return self.val

class AddMarginProduct(nn.Module):
    r"""Implement of large margin cosine distance: :
    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        s: norm of input feature
        m: margin
        cos(theta) - m
    """

    def __init__(self, in_features, out_features, s=30.0, m=0.40):
        super(AddMarginProduct, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.s = s
        self.m = m
        self.weight = Parameter(torch.FloatTensor(out_features, in_features))
        nn.init.xavier_uniform_(self.weight)

    def forward(self, input, label):
        # --------------------------- cos(theta) & phi(theta) ---------------------------
        cosine = F.linear(F.normalize(input), F.normalize(self.weight))
        phi = cosine - self.m
        # --------------------------- convert label to one-hot ---------------------------
        one_hot = torch.zeros(cosine.size(), device='cuda')
        # one_hot = one_hot.cuda() if cosine.is_cuda else one_hot
        one_hot.scatter_(1, label.view(-1, 1).long(), 1)
        # -------------torch.where(out_i = {x_i if condition_i else y_i) -------------
        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)  # you can use torch.where if your torch.__version__ is 0.4
        output *= self.s
        # print(output)

        return output

    def __repr__(self):
        return self.__class__.__name__ + '(' \
               + 'in_features=' + str(self.in_features) \
               + ', out_features=' + str(self.out_features) \
               + ', s=' + str(self.s) \
               + ', m=' + str(self.m) + ')'


class SphereProduct(nn.Module):
    r"""Implement of large margin cosine distance: :
    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        m: margin
        cos(m*theta)
    """
    def __init__(self, in_features, out_features, m=4):
        super(SphereProduct, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.m = m
        self.base = 1000.0
        self.gamma = 0.12
        self.power = 1
        self.LambdaMin = 5.0
        self.iter = 0
        self.weight = Parameter(torch.FloatTensor(out_features, in_features))
        nn.init.xavier_uniform(self.weight)

        # duplication formula
        self.mlambda = [
            lambda x: x ** 0,
            lambda x: x ** 1,
            lambda x: 2 * x ** 2 - 1,
            lambda x: 4 * x ** 3 - 3 * x,
            lambda x: 8 * x ** 4 - 8 * x ** 2 + 1,
            lambda x: 16 * x ** 5 - 20 * x ** 3 + 5 * x
        ]

    def forward(self, input, label):
        # lambda = max(lambda_min,base*(1+gamma*iteration)^(-power))
        self.iter += 1
        self.lamb = max(self.LambdaMin, self.base * (1 + self.gamma * self.iter) ** (-1 * self.power))

        # --------------------------- cos(theta) & phi(theta) ---------------------------
        cos_theta = F.linear(F.normalize(input), F.normalize(self.weight))
        cos_theta = cos_theta.clamp(-1, 1)
        cos_m_theta = self.mlambda[self.m](cos_theta)
        theta = cos_theta.data.acos()
        k = (self.m * theta / 3.14159265).floor()
        phi_theta = ((-1.0) ** k) * cos_m_theta - 2 * k
        NormOfFeature = torch.norm(input, 2, 1)

        # --------------------------- convert label to one-hot ---------------------------
        one_hot = torch.zeros(cos_theta.size())
        one_hot = one_hot.cuda() if cos_theta.is_cuda else one_hot
        one_hot.scatter_(1, label.view(-1, 1), 1)

        # --------------------------- Calculate output ---------------------------
        output = (one_hot * (phi_theta - cos_theta) / (1 + self.lamb)) + cos_theta
        output *= NormOfFeature.view(-1, 1)

        return output

    def __repr__(self):
        return self.__class__.__name__ + '(' \
               + 'in_features=' + str(self.in_features) \
               + ', out_features=' + str(self.out_features) \
               + ', m=' + str(self.m) + ')'

#---------------------------------------------------------------------------
# Metric adaptation for deep supervision

class DeepMetric(Metric):
    def __init__(self,
        metric,
        alphas, # list of coefficient applied to each feature map, starts with the deepest one, must have a len of 6
        name=None,
        metric_kwargs={}):
        super(DeepMetric, self).__init__()
        self.metric = metric(**metric_kwargs)
        self.name = name 
        self.alphas = alphas 
    
    def forward(self, inputs, targets):
        # inputs must be a list of network output
        # they are here all compared to the targets
        # the last inputs is supposed to be the final one
        self.val = 0
        for i in range(len(inputs)):
            if self.alphas[i]!=0:
                self.val += self.metric(inputs[i], targets)*self.alphas[i]
        return self.val

#---------------------------------------------------------------------------



import torch
import os
# from telegram_send import send
import matplotlib.pyplot as plt
from torch.utils.tensorboard import SummaryWriter

import numpy as np

#----------------------------------------------------------------------------
# Base classes

class Callback(object):
    """
    Abstract base class used to build new callbacks.
    """

    def __init__(self):
        self.metrics = None

    def set_trainer(self, metrics):
        self.metrics = metrics

    def on_batch_begin(self, batch):
        pass

    def on_batch_end(self, batch):
        pass

    def on_epoch_begin(self, epoch):
        pass

    def on_epoch_end(self, epoch):
        pass

    def on_train_begin(self):
        pass

    def on_train_end(self):
        pass

# dict based callbacks

class Callbacks(Callback):
    def __init__(self, callbacks):
        super().__init__()
        if isinstance(callbacks, Callbacks):
            callbacks = callbacks.callbacks
        self.callbacks = callbacks
        if callbacks is None:
            self.callbacks = {}
    
    def __getitem__(self, name): return self.callbacks[name]

    def set_trainer(self, trainer):
        for callback in self.callbacks.values():
            callback.set_trainer(trainer)

    def on_batch_begin(self, batch):
        for callback in self.callbacks.values():
            callback.on_batch_begin(batch)

    def on_batch_end(self, batch):
        for callback in self.callbacks.values():
            callback.on_batch_end(batch)

    def on_epoch_begin(self, epoch):
        for callback in self.callbacks.values():
            callback.on_epoch_begin(epoch)

    def on_epoch_end(self, epoch):
        for callback in self.callbacks.values():
            callback.on_epoch_end(epoch)

    def on_train_begin(self):
        for callback in self.callbacks.values():
            callback.on_train_begin()

    def on_train_end(self):
        for callback in self.callbacks.values():
            callback.on_train_end()

#----------------------------------------------------------------------------
# Savers            
            
class ModelSaver(Callback): # TODO: save best_only
    """
    if model is a list then it is considered as been [student,teacher]
    """
    def __init__(self,
        model, 
        optimizer, 
        path='unet', 
        every_epoch=2, 
        save_best=True, 
        loss=None, # loss used for saving the best model, generally the val loss
        saved_loss=None, # loss being saved
        ):

        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.path_last = path + '.pth'
        self.path_best = path + '_best.pth'
        self.every_epoch = every_epoch
        self.save_best = save_best
        self.best_loss = float('inf')
        self.loss = loss
        self.saved_loss = saved_loss

    def on_epoch_end(self, epoch):
        if isinstance(self.model, list):
            save_dict = {
                'epoch':   epoch,
                'student': self.model[0].state_dict(), 
                'teacher': self.model[1].state_dict(), 
                'opt':     self.optimizer.state_dict(),
                'loss':    self.saved_loss.state_dict(),
                }
        else: 
            save_dict = {
                    'epoch':    epoch,
                    'model':    self.model.state_dict(), 
                    'opt':      self.optimizer.state_dict(),
                    'loss':     self.saved_loss.state_dict(),
                    }
        if epoch % self.every_epoch == 0:
            torch.save(save_dict, self.path_last)
            print('Save model to {}'.format(self.path_last))

        # save best model if needed 
        if self.save_best and (self.loss.avg < self.best_loss) and (self.loss.avg != 0):
            torch.save(save_dict, self.path_best)
            print('Save best model to {}'.format(self.path_best))
            self.best_loss = self.loss.avg
            
                

class LogSaver(Callback):
    def __init__(self, 
        log_dir,            # path to where the csv file will be store
        train_loss,
        val_loss,
        train_metrics,
        val_metrics,
        scheduler=None,
        save_best=False,    # save the best metrics in another text file ? 
        every_batch=10):    # save period (in batch)

        self.path = log_dir + '/log.csv'

        self.crt_epoch = 1

        self.train_loss = train_loss
        self.val_loss = val_loss
        self.train_metrics = train_metrics
        self.val_metrics = val_metrics

        self.scheduler = scheduler

        if save_best:
            self.save_best = save_best # self.tracked_loss = val_loss
            self.best_loss = float('inf')
            self.best_path = log_dir + '/log_best.csv'
            self.f_best = None

        self.every_batch = every_batch

        self.f = None

    def write_file_head(self, file):
        # write the head of the log file
        head = "epoch,batch,learning_rate,train_loss,val_loss"
        for m in self.train_metrics+self.val_metrics:
            head += "," + m.name
        file.write(head + "\n")

    def on_train_begin(self):
        # create the file and edit the head 
        self.f = open(self.path, "a")
        if os.stat(self.path).st_size == 0: # if the file is empty
            self.write_file_head(self.f)

    def on_epoch_begin(self, epoch):
        self.crt_epoch = epoch+1

    def on_batch_end(self, batch):
        if batch % self.every_batch == 0:
            template =  str(self.crt_epoch)
            template += "," + str(batch)
            template += "," + str(self.scheduler.get_last_lr()[0])
            template += "," + str(self.train_loss.val.item()) # TODO: save the avg value only
            # BUG fix below: TODO, simplify it! 
            val_loss = self.val_loss.val if type(self.val_loss.val)==int else self.val_loss.val.item()
            template += "," + str(val_loss)
            for m in self.train_metrics: template += "," + str(m.val.item())
            for m in self.val_metrics: 
                # TODO: cf. previous comment
                val_m = m.avg if type(m.avg)==int else m.avg.item() 
                template += "," + str(val_m)
            self.f.write(template + "\n")
        
            # save best if needed
            if self.save_best \
                and (self.best_loss > self.val_loss.val) \
                and (self.val_loss.avg != 0): # to avoid saving during the first epoch
                self.best_loss = self.val_loss.avg
                self.f_best = open(self.best_path, "w") # open it in write mode
                self.write_file_head(self.f_best)
                self.f_best.write(template)
                self.f_best.close()

# for image saver TODO:
# use the run_prediction_single method of the Builder class itself
# EXAMPLE:
# class test:
#     def __init__(self):
#         self.a = 0
#     def t(self):
#         self.a += 1
#     def __call__(self):
#         caller(self, self.t)

# def caller(tester, fct):
#     print(tester)
#     print(tester.a)
#     fct()
#     print(tester.a)

# tester = test()
# tester()

class ImageSaver(Callback):
    def __init__(self, 
        image_dir, 
        model,
        val_dataloader,
        use_sigmoid=True,
        every_epoch=1):
        self.image_dir = image_dir
        self.every_epoch = every_epoch
        self.model = model 
        self.use_sigmoid = use_sigmoid
        self.val_dataloader = val_dataloader

    def on_epoch_end(self, epoch):
        if epoch % self.every_epoch == 0:
            self.model.eval()
            with torch.no_grad():
                # for _ in range(4): next(iterator) # to start with image number 4
                plot_size = 1
                for i in range(plot_size):
                # make prediction
                    X, y = next(iter(self.val_dataloader))
                    # X, y = self.val_dataloader.get_sample()
                    X, y = X.cuda(), y.cuda()
                    with torch.cuda.amp.autocast():
                        pred = self.model(X).detach()
                    if self.use_sigmoid:
                        pred = (torch.sigmoid(pred)>0.5).int()*255
                    else: 
                        pred = (pred.softmax(dim=1)).int()*255
                    l = [X, y, pred]
                    for j in range(len(l)):
                        _,_,channel,_,_ = l[j].shape
                        l[j] = l[j][0, 0, channel//2, ...].cpu().numpy().astype(float)
                    X, y, pred = l

                    # plot 
                    # plt.figure(figsize=(15,5))
                    plt.figure(dpi=100)
                    # print original
                    plt.subplot(plot_size,3,3*i+1)
                    plt.imshow(X)
                    plt.title('raw')
                    plt.axis('off')
                    # print prediction
                    plt.subplot(plot_size,3,3*i+2)
                    plt.imshow(pred)
                    plt.title('pred')
                    plt.axis('off')
                    # print ground truth
                    plt.subplot(plot_size,3,3*i+3)
                    plt.imshow(y)
                    plt.title('gt')
                    plt.axis('off')
                    del X, y, pred
                im_path = os.path.join(self.image_dir,'image_' + str(epoch) + '.png')
                print("Save image to {}".format(im_path))
                plt.savefig(im_path)
                plt.close()

class TensorboardSaver(Callback):
    def __init__(self, 
        log_dir, 
        train_loss, 
        val_loss, 
        train_metrics, 
        val_metrics, 
        batch_size,
        n_batch_per_epoch):

        self.writer = SummaryWriter(log_dir=log_dir)
        self.train_loss = train_loss
        self.val_loss = val_loss
        self.train_metrics = train_metrics
        self.val_metrics = val_metrics
        self.batch_size = batch_size
        self.n_batch_per_epoch = n_batch_per_epoch

        self.crt_epoch = 0
    
    def on_epoch_begin(self, epoch):
        self.crt_epoch = epoch
    
    def on_batch_end(self, batch):
        n_iter = (self.n_batch_per_epoch * self.crt_epoch + batch) * self.batch_size
        self.writer.add_scalar('Loss/train', self.train_loss.val, n_iter)
        if self.val_loss:
            self.writer.add_scalar('Loss/test', self.val_loss.avg, n_iter)
        if self.train_metrics:
            for m in self.train_metrics: self.writer.add_scalar('Metrics/'+m.name,m.val,n_iter)
        if self.val_metrics:
            for m in self.val_metrics: self.writer.add_scalar('Metrics/'+m.name,m.avg,n_iter)

#----------------------------------------------------------------------------
# Printer

class LogPrinter(Callback):
    def __init__(self, metrics, nbof_epochs, nbof_batches, every_batch=10):
        self.nbof_epochs = nbof_epochs
        self.nbof_batches = nbof_batches

        self.metrics = metrics

        self.every_batch = every_batch

    def on_epoch_begin(self, epoch):
        print("Epoch [{:>3d}/{:>3d}]".format(epoch+1, self.nbof_epochs))

    def on_batch_end(self, batch):
        if batch % self.every_batch == 0:
            template = "Batch [{:>3d}/{:>3d}]".format(batch, self.nbof_batches)
            for i in range(len(self.metrics)):
                template += ", {}".format(self.metrics[i])
            print(template)

# class Telegram(Callback):
#     """
#     Send message when training finishes
#     """
#     def __init__(self, loss, test_loss=None):
#         self.loss = loss
#         self.test_loss = test_loss
#         self.template = "Finished training with loss {} and test loss {}"
    
#     def on_train_end(self):
#         send(messages=[self.template.format(self.loss.val, self.test_loss.avg)])

#----------------------------------------------------------------------------
# schedulers

class LRSchedulerMultiStep(Callback):
    """
    Multi-step scheduler only for now
    """
    def __init__(self, optimizer, milestones, gamma=0.1):
        self.scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones, gamma=gamma, verbose=False)

    def get_last_lr(self):
        return self.scheduler.get_last_lr()
    
    def on_epoch_end(self, epoch):
        self.scheduler.step()
        print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class LRSchedulerCosine(Callback):
    """
    Cosine scheduler
    """
    def __init__(self, optimizer, T_max):
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max, eta_min=1e-6, verbose=False)

    def get_last_lr(self):
        return self.scheduler.get_last_lr()
    
    def on_epoch_end(self, epoch):
        self.scheduler.step()
        print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class LRSchedulerPoly(Callback):
    """
    Polygonal scheduler
    """
    def __init__(self, optimizer, initial_lr, max_epochs, exponent=0.9):
        self.initial_lr = initial_lr
        self.max_epochs = max_epochs
        self.exponent = exponent
        self.optimizer = optimizer

    # def get_last_lr(self):
    #     return self.optimizer.param_groups[0]['lr']
    
    def on_train_begin(self):
        self.optimizer.param_groups[0]['lr'] = self.initial_lr
        print("Current learning rate: {}".format(self.optimizer.param_groups[0]['lr']))

    def on_epoch_end(self, epoch):
        self.optimizer.param_groups[0]['lr'] = self.initial_lr * (1 - epoch / self.max_epochs)**self.exponent
        print("Current learning rate: {}".format(self.optimizer.param_groups[0]['lr']))


# class LRSchedulerCosine(Callback):
#     """
#     Multi-step scheduler only for now
#     """
#     def __init__(self, optimizer, T_max):
#         self.scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=50, T_mult=2, eta_min=1e-6, last_epoch=-1, verbose=False)

#     def get_last_lr(self):
#         return self.scheduler.get_last_lr()
    
#     def on_epoch_end(self, epoch):
#         self.scheduler.step()
#         print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class ForceFGScheduler(Callback):
    """
    Force foreground scheduler
    We force the model to "see" the foreground more often in the beginning and progressively reduce it
    This function edits the dataloader.dataset.fg_rate argument --> TODO: improve that
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    def on_train_begin(self):
        self.dataloader.dataset.set_fg_rate(self.initial_rate)
        print("Current foreground rate: {}".format(self.initial_rate))
    
    def on_epoch_end(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_fg_rate(crt_rate) 
        print("Current foreground rate: {}".format(crt_rate))

class OverlapScheduler(Callback):
    """
    Overlap scheduler
    We progressively reduce the minimum overlap between the global patches and the local patches
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    def on_train_begin(self):
        self.dataloader.dataset.set_min_overlap(self.initial_rate)
        print("Current overlap: {}".format(self.initial_rate))
    
    def on_epoch_end(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_min_overlap(crt_rate)
        print("Current overlap: {}".format(crt_rate))

class GlobalScaleScheduler(Callback):
    """
    Global scale scheduler
    We progressively reduce the scale of the global_crop from image size to patch size
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    def on_train_begin(self):
        self.dataloader.dataset.set_global_crop(self.initial_rate)
        print("Current global crop scale: {}".format(self.initial_rate))
    
    def on_epoch_end(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_global_crop(crt_rate)
        print("Current global crop scale: {}".format(crt_rate))

class WeightDecayScheduler(Callback):
    """
    cosine scheduler of the weight decay
    if use_poly=True, use polynomial update instead of cosine
    """
    def __init__(self, optimizer, initial_wd, final_wd, nb_epochs, use_poly=False, exponent=0.9):
        self.optimizer = optimizer
        self.initial_wd = initial_wd
        self.final_wd = final_wd
        self.nb_epochs = nb_epochs
        self.use_poly = use_poly 
        self.exponent = exponent

    def on_train_begin(self):
        self.optimizer.param_groups[0]["weight_decay"] = self.initial_wd
        print("Current weight decay:", self.optimizer.param_groups[0]["weight_decay"] )
    
    def on_epoch_end(self, epoch):
        if self.use_poly:
            self.optimizer.param_groups[0]["weight_decay"] = self.final_wd + (self.initial_wd - self.final_wd) * (1 - epoch / self.nb_epochs)**self.exponent
        else: 
            self.optimizer.param_groups[0]["weight_decay"] = self.final_wd + 0.5 * (self.initial_wd - self.final_wd) * (1 + np.cos(np.pi * epoch / self.nb_epochs))
        print("Current weight decay:", self.optimizer.param_groups[0]["weight_decay"] )

class MomentumScheduler(Callback):
    """
    cosine scheduler for the momentum of the teacher model update
    if use_poly=True, use polynomial update instead of cosine
    """
    def __init__(self, initial_momentum, final_momentum, nb_epochs, mode='poly', exponent=0.9):
        self.initial_momentum = initial_momentum
        self.final_momentum = final_momentum
        self.nb_epochs = nb_epochs
        self.crt_momentum = None
        self.mode = mode 
        self.exponent = exponent
    
    def __getitem__(self, epoch):
        if self.mode=='poly':
            self.crt_momentum = self.final_momentum + (self.initial_momentum - self.final_momentum) * (1 - epoch / self.nb_epochs)**self.exponent
        elif self.mode=='exp':
            self.crt_momentum = self.final_momentum + (self.initial_momentum - self.final_momentum) * (np.exp(-self.exponent*epoch / self.nb_epochs))
        elif self.mode=='linear':
            if epoch > self.nb_epochs:
                self.crt_momentum = self.final_momentum
            else:
                self.crt_momentum = self.initial_momentum + (self.final_momentum - self.initial_momentum) * epoch / self.nb_epochs
        # elif self.mode=='mix': # linear increase and then an exponential increase
        #     if epoch > self.nb_epochs:
        #         self.crt_momentum = self.final_momentum
        #     else:
        #         self.crt_momentum = self.initial_momentum + (self.final_momentum - self.initial_momentum) * epoch / self.nb_epochs
        else: # cosine
            self.crt_momentum = self.final_momentum + 0.5 * (self.initial_momentum - self.final_momentum) * (1 + np.cos(np.pi * epoch / self.nb_epochs))
        return self.crt_momentum

    def on_epoch_end(self, epoch):
        print("Current teacher momentum:", self.crt_momentum)

class DatasetSizeScheduler(Callback):
    """
    Dataset size scheduler
    We progressively increase the size of the dataset to help the arcface training
    """
    def __init__(self, dataloader, model, max_dataset_size, min_dataset_size=5):
        self.dataloader = dataloader
        self.model = model
        self.max_dataset_size = max_dataset_size
        self.min_dataset_size = min_dataset_size

    def on_epoch_begin(self, epoch):
        dataset_size = min(epoch+self.min_dataset_size, self.max_dataset_size)
        self.dataloader.dataset.set_dataset_size(dataset_size)
        self.model.set_num_classes(dataset_size)
        print("Current dataset size: {}".format(dataset_size))

#----------------------------------------------------------------------------
# metrics updater

class MetricsUpdater(Callback):
    """
    Update the metrics averages
    """
    def __init__(self, metrics, batch_size):
        self.metrics = metrics
        self.batch_size = batch_size

    def on_epoch_begin(self, epoch=None):
        for m in self.metrics: m.reset()

    def on_batch_end(self, batch=None):
        for m in self.metrics: m.update(self.batch_size)

#----------------------------------------------------------------------------


#---------------------------------------------------------------------------
# Dataset primitives for 3D segmentation dataset
# solution: patch approach with the whole dataset into memory 
#---------------------------------------------------------------------------

from json import load
import os
import numpy as np 
import torchio as tio
import random 
import torch
from torch.utils.data import Dataset, DataLoader
# from monai.data import CacheDataset
import pandas as pd 
from tqdm import tqdm 
from skimage.io import imread

from biom3d.utils import centered_pad, get_folds_train_test_df

#---------------------------------------------------------------------------
# utilities to random crops
    
# create a patch centered on this voxel
def centered_crop(img, msk, center, crop_shape):
    """
    centered crop a portion of size prop of the original image size.
    """
    crop_shape=np.array(crop_shape)
    start = center-crop_shape//2
    end = crop_shape+start

    crop_img = img[:,
                start[0]:end[0], 
                start[1]:end[1], 
                start[2]:end[2]]
    crop_msk = msk[:,
                start[0]:end[0], 
                start[1]:end[1], 
                start[2]:end[2]]
    return crop_img, crop_msk

def foreground_crop(img, msk, foreground, crop_shape, margin=np.zeros(3)):
    """
    crop the img and msk so that the foreground voxel is located somewhere (random) in the cropped patch
    the margin argument adds a margin to force the foreground voxel to be located nearer the middle of the patch 
    """
    lower_bound = np.maximum(0,foreground-np.array(crop_shape)+margin)
    start = np.random.randint(low=lower_bound, high=np.maximum(lower_bound+1,foreground-margin))
    end = start+crop_shape
    
    crop_img = img[:,
                start[0]:end[0], 
                start[1]:end[1], 
                start[2]:end[2]]
    crop_msk = msk[:,
                start[0]:end[0], 
                start[1]:end[1], 
                start[2]:end[2]]
    return crop_img, crop_msk

def random_crop(img, msk, crop_shape):
    """
    randomly crop a portion of size prop of the original image size.
    """
    img_shape = np.array(img.shape)[1:]
    # rand_start = np.array([random.randint(0,c) for c in np.maximum(0,(img_shape-crop_shape))])
    rand_start = np.random.randint(0, np.maximum(1,img_shape-crop_shape))
    rand_end = crop_shape+rand_start

    crop_img = img[:,
                    rand_start[0]:rand_end[0], 
                    rand_start[1]:rand_end[1], 
                    rand_start[2]:rand_end[2]]
    crop_msk = msk[:,
                    rand_start[0]:rand_end[0], 
                    rand_start[1]:rand_end[1], 
                    rand_start[2]:rand_end[2]]
    return crop_img, crop_msk

# def centered_pad(img, msk, final_size):
#     """
#     centered pad an img and msk to fit the final_size
#     """
#     final_size = np.array(final_size)
#     img_shape = np.array(img.shape[1:])
    
#     start = (final_size-np.array(img_shape))//2
#     end = final_size-(img_shape+start)
#     end = end * (end > 0)
    
#     pad = np.append([[0,0]], np.stack((start,end),axis=1), axis=0)
#     pad_img = np.pad(img, pad, 'constant', constant_values=0)
#     pad_msk = np.pad(msk, pad, 'constant', constant_values=0)
    
# #     if ((final_size-np.array(pad_img.shape)) < 0).any(): # keeps only the negative values
# #         pad_img = pad_img[:,:final_size[0],:final_size[1],:final_size[2]]
# #         pad_msk = pad_msk[:,:final_size[0],:final_size[1],:final_size[2]]
#     return pad_img, pad_msk

# def random_pad(img, msk, final_size):
#     """
#     [CAREFUL!] I THINK THIS FUNCTION HAS SOME BUGS (WITH SMALL IMAGES)
#     randomly pad an image with zeros to reach the final size. 
#     if the image is bigger than the expected size, then the image is cropped.
#     """
#     img_shape = np.array(img.shape)[1:]
#     size_range = (final_size-img_shape) * (final_size-img_shape > 0) # needed if the original image is bigger than the final one
#     # rand_start = np.array([random.randint(0,c) for c in size_range])
#     rand_start = np.random.randint(0,np.maximum(1,size_range))

#     rand_end = final_size-(img_shape+rand_start)
#     rand_end = rand_end * (rand_end > 0)

#     pad = np.append([[0,0]],np.vstack((rand_start, rand_end)).T,axis=0)
#     pad_img = np.pad(img, pad, 'constant', constant_values=0)
#     pad_msk = np.pad(msk, pad, 'constant', constant_values=0)
#     # pad_img = torch.nn.functional.pad(img, tuple(pad.flatten().tolist()), 'constant', value=0)

#     # crop the image if needed
#     if ((final_size-np.array(pad_img.shape)) < 0).any(): # keeps only the negative values
#         pad_img = pad_img[:,:final_size[0],:final_size[1],:final_size[2]]
#         pad_msk = pad_msk[:,:final_size[0],:final_size[1],:final_size[2]]
#     return pad_img, pad_msk

def random_crop_pad(img, msk, final_size, fg_rate=0.33, fg_margin=np.zeros(3), remove_bg=False):
    """
    random crop and pad if needed.
    """
    # choose if using foreground centrered or random alignement
    force_fg = random.random()
    if fg_rate>0 and force_fg<fg_rate:
        start_rnd = 1 if remove_bg else 0
        rnd_label = random.randint(start_rnd,msk.shape[0]-1) # choose a random label
        
        locations = np.argwhere(msk[rnd_label] == 1)
        
        if locations.size==0: # bug fix when having empty arrays 
            img, msk = random_crop(img, msk, final_size)
        else:
            center=random.choice(locations) # choose a random voxel of this label
            img, msk = foreground_crop(img, msk, center, final_size, fg_margin)
    else:
        # or random crop
        img, msk = random_crop(img, msk, final_size)
        
    # pad if needed
    if np.any(np.array(img.shape)[1:]-final_size)!=0:
        img, msk = centered_pad(img=img, msk=msk, final_size=final_size)
    return img, msk

def random_crop_resize(img, msk, crop_scale, final_size, fg_rate=0.33, fg_margin=np.zeros(3), remove_bg=False):
    """
    random crop and resize if needed.
    Args:
    crop_scale: >=1
    """
    final_size = np.array(final_size)
        
    # determine crop shape
    max_crop_shape = np.round(final_size * crop_scale).astype(int)
    crop_shape = np.random.randint(final_size, max_crop_shape+1)
    
    # choose if using foreground centrered or random alignement
    force_fg = random.random()
    if fg_rate>0 and force_fg<fg_rate:
        start_rnd = 1 if remove_bg else 0
        rnd_label = random.randint(start_rnd,msk.shape[0]-1) # choose a random label
        
        locations = np.argwhere(msk[rnd_label] == 1)
        
        if locations.size==0: # bug fix when having empty arrays 
            img, msk = random_crop(img, msk, crop_shape)
        else:
            center=random.choice(locations) # choose a random voxel of this label
            img, msk = foreground_crop(img, msk, center, crop_shape, fg_margin)
    else:
        # or random crop
        img, msk = random_crop(img, msk, crop_shape)
    
    # resize if needed
    if np.any(np.array(img.shape)[1:]-final_size)>0:
        sub = tio.Subject(img=tio.ScalarImage(tensor=img), msk=tio.LabelMap(tensor=msk))
        sub = tio.Resize(final_size)(sub)
        img, msk = sub.img.tensor, sub.msk.tensor
    elif np.any(np.array(img.shape)[1:]-final_size)<0:
        img, msk = centered_pad(img, msk, final_size)
    return img, msk

#---------------------------------------------------------------------------

class SemSeg3DPatchFast(Dataset):
    """
    with DataLoader
    """
    def __init__(
        self,
        img_dir,
        msk_dir,
        batch_size, 
        patch_size,
        nbof_steps,
        folds_csv  = None, 
        fold       = 0, 
        val_split  = 0.25,
        train      = True,
        use_aug    = True,
        aug_patch_size = None,
        fg_rate = 0.33, # if > 0, force the use of foreground, needs to run some pre-computations (note: better use the foreground scheduler)
        crop_scale = 1.0, # if > 1, then use random_crop_resize instead of random_crop_pad
        use_softmax = True, # if true, means that the output is one_hot encoded for softmax use
        load_data = False, # if True, loads the all dataset into computer memory (faster but more memory expensive)
        ):

        self.img_dir = img_dir
        self.msk_dir = msk_dir

        self.batch_size = batch_size
        self.patch_size = patch_size
        self.aug_patch_size = aug_patch_size

        self.nbof_steps = nbof_steps

        self.load_data = load_data
        
        # get the training and validation names 
        if folds_csv is not None:
            df = pd.read_csv(folds_csv)
            trainset, testset = get_folds_train_test_df(df, verbose=False)

            self.fold = fold
            
            self.val_imgs = trainset[self.fold]
            del trainset[self.fold]
            self.train_imgs = []
            for i in trainset: self.train_imgs += i

        else: # tmp: validation split = 50% by default
            all_set = os.listdir(img_dir)
            val_split = np.round(val_split * len(all_set)).astype(int)
            if val_split == 0: val_split=1
            self.train_imgs = all_set[val_split:]
            self.val_imgs = all_set[:val_split]
            testset = []
        
        if self.load_data:
            print("Loading the whole dataset into computer memory...")
            def load_data(imgs_fnames):
                imgs_data = []
                msks_data = []
                for idx in range(len(imgs_fnames)):
                    # file names
                    img_path = os.path.join(self.img_dir, imgs_fnames[idx])
                    msk_path = os.path.join(self.msk_dir, imgs_fnames[idx])

                    # load img and msks
                    imgs_data += [imread(img_path)]
                    msks_data += [imread(msk_path)]
                return imgs_data, msks_data

            self.train_imgs_data, self.train_msks_data = load_data(self.train_imgs)
            self.val_imgs_data, self.val_msks_data = load_data(self.val_imgs)
            print("Done!")
            
        self.train = train
        print("current fold: {}\n \
        length of the training set: {}\n \
        length of the validation set: {}\n \
        length of the testing set: {}\n \
        is training mode active?: {}".format(fold, len(self.train_imgs), len(self.val_imgs), len(testset), self.train))

        # print train and validation image names
        print("Training images:", self.train_imgs)
        print("Validation images:", self.val_imgs)

        self.use_aug = use_aug

        if self.use_aug:
            ps = np.array(self.patch_size)

            # [aug] flipping probabilities
            flip_prop=ps.min()/ps
            flip_prop/=flip_prop.sum()

            # [aug] 'axes' for tio.RandomAnisotropy
            anisotropy_axes=tuple(np.arange(3)[ps/ps.min()>3].tolist())

            # [aug] 'degrees' for tio.RandomAffine
            if np.any(ps/ps.min()>3): # then use dummy_2d
                norm = ps*3/ps.min()
                softmax=np.exp(norm)/sum(np.exp(norm))
                degrees=softmax.min()*90/softmax
                # degrees = 180*ps.min()/ps
            else:
                degrees = (-30,30)

            # [aug] 'cropping'
            # the affine transform is computed on bigger patches than the other transform
            # that's why we need to crop the patch after potential affine augmentation
            start = (np.array(self.aug_patch_size)-np.array(self.patch_size))//2
            end = self.aug_patch_size-(np.array(self.patch_size)+start)
            cropping = (start[0],end[0],start[1],end[1],start[2],end[2])
            
            # the foreground-crop-function forces the foreground to be in the center of the patch
            # so that, when doing the second centrering crop, the foreground is still present in the patch,
            # that's why there is a margin here
            self.fg_margin = start 


            self.transform = tio.Compose([
                # spatial augmentations
                # tio.RandomFlip(p=0.8, axes=(0), flip_probability=flip_prop[0]),
                # tio.RandomFlip(p=0.8, axes=(1), flip_probability=flip_prop[1]),
                # tio.RandomFlip(p=0.8, axes=(2), flip_probability=flip_prop[2]),
                # tio.RandomFlip(p=0.2, axes=(0,1,2)),
                # tio.RandomAnisotropy(p=0.25, axes=anisotropy_axes, downsampling=(1,2)),
                tio.RandomAffine(p=0.25, scales=(0.7,1.4), degrees=degrees, translation=0),
                tio.Crop(p=1, cropping=cropping),

                tio.RandomFlip(p=1, axes=(0,1,2)),
                # tio.RandomElasticDeformation(p=0.2, num_control_points=4, locked_borders=1),
                # tio.OneOf({
                #     tio.RandomAffine(scales=0.1, degrees=10, translation=0): 0.8,
                #     tio.RandomElasticDeformation(): 0.2,
                # }),
                

                # intensity augmentations
                # tio.RandomMotion(p=0.2),
                # tio.RandomGhosting(p=0.2),
                # tio.RandomSpike(p=0.15),
                tio.RandomBiasField(p=0.15, coefficients=0.2),
                tio.RandomBlur(p=0.2, std=(0.5,1.5)),
                tio.RandomNoise(p=0.2, std=(0,0.1)),
                # tio.RandomSwap(p=0.2, patch_size=ps//8),
                tio.RandomGamma(p=0.3, log_gamma=(-0.35,0.4)),
            ])

        # self.fg_rate = fg_rate if self.train else 1
        self.fg_rate = fg_rate
        self.use_softmax = use_softmax
        self.crop_scale = crop_scale 
        assert self.crop_scale >= 1, "[Error] crop_scale must be higher or equalt to 1"
    
    def set_fg_rate(self,value):
        """
        setter function for the foreground rate class parameter
        """
        self.fg_rate = value
    
    def __len__(self):
        # return len(self.train_imgs) if self.training else len(self.val_imgs)
        return self.nbof_steps*self.batch_size
    
    def __getitem__(self, idx):

        if self.load_data:
            if self.train:
                img = self.train_imgs_data[idx%len(self.train_imgs_data)]
                msk = self.train_msks_data[idx%len(self.train_msks_data)]
            else: 
                img = self.val_imgs_data[idx%len(self.val_imgs_data)]
                msk = self.val_msks_data[idx%len(self.val_msks_data)]
        else:
            fnames = self.train_imgs if self.train else self.val_imgs
            # img_fname = np.random.choice(fnames)
            img_fname = fnames[idx%len(fnames)]
            
            # file names
            img_path = os.path.join(self.img_dir, img_fname)
            msk_path = os.path.join(self.msk_dir, img_fname)

            # read the images
            img = imread(img_path)
            msk = imread(msk_path)

        # remove bg channel if use_softmax is False
        if not self.use_softmax:
            msk = msk[1:]

        # random crop and pad
        final_size = self.aug_patch_size if self.use_aug else self.patch_size
        fg_margin = self.fg_margin if self.use_aug else np.zeros(3)
        if self.train and self.crop_scale > 1:
            img, msk = random_crop_resize(
                img,
                msk,
                crop_scale=self.crop_scale,
                final_size=final_size,
                fg_rate=self.fg_rate,
                fg_margin=fg_margin,
                remove_bg=self.use_softmax,
                )
        else:
            img, msk = random_crop_pad(
                img,
                msk,
                final_size=final_size,
                fg_rate=self.fg_rate,
                fg_margin=fg_margin,
                remove_bg=self.use_softmax,
                )

        # data augmentation
        if self.use_aug:
            sub = tio.Subject(img=tio.ScalarImage(tensor=img), msk=tio.LabelMap(tensor=msk))
            sub = self.transform(sub)
            img, msk = sub.img.tensor, sub.msk.tensor
        
            # to float for msk
            msk = msk.float()

        elif not self.use_softmax:
            msk = msk.astype(float)
        
        return img, msk

#---------------------------------------------------------------------------


import argparse
import os
from omero.cli import cli_login
from omero.gateway import BlitzGateway

from biom3d import omero_downloader 
from biom3d import pred 


def run(obj, target, bui_dir, dir_out, host=None, user=None, pwd=None):
    print("Start dataset/project downloading...")
    if host:
        conn = BlitzGateway(user, pwd, host=host, port=4064)
        conn.connect()
        datasets, dir_in = omero_downloader.download_object(conn, obj, target)
        conn.close()
    else:
        with cli_login() as cli:
            datasets, dir_in = omero_downloader.download_object_cli(cli, obj, target)
    print("Done downloading dataset/project!")

    print("Start prediction...")
    if 'Dataset' in obj:
        dir_in = os.path.join(dir_in, datasets[0].name)
        dir_out = os.path.join(dir_out, datasets[0].name)
        if not os.path.isdir(dir_out):
            os.makedirs(dir_out, exist_ok=True)
        pred.pred(bui_dir, dir_in, dir_out)
    elif 'Project' in obj:
        dir_out = os.path.join(dir_out, os.path.split(dir_in)[-1])
        if not os.path.isdir(dir_out):
            os.makedirs(dir_out, exist_ok=True)
        pred.pred_multiple(bui_dir, dir_in, dir_out)
    else:
        print("[Error] Type of object unknown {}. It should be 'Dataset' or 'Project'".format(obj))
    print("Done prediction!")


if __name__=='__main__':

    # parser
    parser = argparse.ArgumentParser(description="Main training file.")
    parser.add_argument('--obj', type=str,
        help="Download object: 'Project:ID' or 'Dataset:ID'")
    parser.add_argument('--target', type=str, default="data/to_pred/",
        help="Directory name to download into")
    parser.add_argument("--bui_dir", type=str, default="logs/unet_nucleus",
        help="Path of the builder directory")
    parser.add_argument("--dir_out", type=str, default="data/pred/",
        help="Path to the output prediction directory")
    parser.add_argument('--hostname', type=str, 
        help="(optional) Host name for Omero server. If not mentioned use the CLI.")
    parser.add_argument('--username', type=str, 
        help="(optional) User name for Omero server")
    parser.add_argument('--password', type=str, 
        help="(optional) Password for Omero server")
    # parser.add_argument("-e", "--eval_only", default=False,  action='store_true', dest='eval_only',
    #     help="Do only the evaluation and skip the prediction (predictions must have been done already.)") 
    args = parser.parse_args()

    run(
        obj=args.obj,
        target=args.target,
        bui_dir=args.bui_dir,
        dir_out=args.dir_out,
        host=args.hostname,
        user=args.username,
        pwd=args.password,
    )


#---------------------------------------------------------------------------
# Dataset preparation to fasten the training
#   -normalization
#   -expand dims and one_hot encoding
#   -saving to tif file
#---------------------------------------------------------------------------

import numpy as np
import os 
# import torchio as tio
from tqdm import tqdm
from skimage.io import imsave, imread
import argparse
import SimpleITK as sitk

#---------------------------------------------------------------------------
# Nifti imread

def sitk_imread(img_path):
    """
    image reader for nii.gz files
    """
    img = sitk.ReadImage(img_path)
    img_np = sitk.GetArrayFromImage(img)
    return img_np, np.array(img.GetSpacing())

def adaptive_imread(img_path):
    """
    use skimage imread or sitk imread depending on the file extension:
    .tif --> skimage.io.imread
    .nii.gz --> SimpleITK.imread
    """
    extension = img_path[img_path.rfind('.'):]
    if extension == ".tif":
        return imread(img_path), []
    else:
        return sitk_imread(img_path)

#---------------------------------------------------------------------------
# Nifti convertion (Medical segmentation decathlon)
# normalization: z-score
# one_hot encoding
# no resampling yet

def one_hot(values, num_classes=None):
    """
    transform the values np.array into a one_hot encoded
    """
    if num_classes==None: n_values = np.max(values) + 1
    else: n_values = num_classes
    # WARNING! potential bug if we have 255 label
    # this function normalize the values to 0,1 if it founds that the maximum of the values if 255
    if values.max()==255: values = (values / 255).astype(np.int64) 
    out = np.eye(n_values)[values]
    return np.moveaxis(out, -1, 0).astype(np.int64)

class Preprocessing:
    """
    A helper class to transform nifti (.nii.gz) and Tiff (.tif or .tiff) images to .tif format and to normalize them.
    """
    def __init__(
        self,
        img_dir,
        img_outdir = None,
        msk_dir = None, # if None, only images are preprocesses not the masks
        msk_outdir = None,
        num_classes = None, # just for debug when empty masks are provided
        remove_bg = True, # keep the background in labels 
        median_spacing=[],
        clipping_bounds=[],
        intensity_moments=[],
        use_tif=True, # use tif instead of npy 
        ):
        """
        Parameters
        ----------
        img_dir : str
            Path to the input image folder
        img_outdir : str
            Path to the output image folder
        msk_dir : str, optional
            Path to the input mask folder
        msk_outdir : str, optional
            Path to the output mask folder
        num_classes : int, optional
            Number of classes (channel) in the masks. Required by the 
        remove_bg : bool, default=True
            Whether to remove the background in the one-hot encoded mask. Remove the background is done when training with sigmoid activations instead of softmax.
        median_spacing : list, optional
            A list of length 3 containing the median spacing of the input images. Median_spacing must not be transposed: for example, median_spacing might be [0.8, 0.8, 2.5] if median shape of the training image is [40,224,224].
        clipping_bounds : list, optional
            A list of length 2 containing the intensity clipping boundary. In nnUNet implementation it corresponds to the 0.5 an 99.5 percentile of the intensities of the voxels of the training images located inside the masks regions.
        intensity_moments : list, optional
            Mean and variance of the intensity of the images voxels in the masks regions. This value are used to normalize the image. 
        """
        assert img_dir!='', "[Error] img_dir must not be empty."

        # fix bug path/folder/ to path/folder
        if os.path.basename(img_dir)=='':
            img_dir = os.path.dirname(img_dir)
        if msk_dir is not None and os.path.basename(msk_dir)=='':
            msk_dir = os.path.dirname(msk_dir)
        
        self.img_dir=img_dir
        self.msk_dir=msk_dir
        self.img_fnames=os.listdir(self.img_dir)

        if img_outdir is None: # name the out dir the same way as the input and add the _out suffix
            img_outdir = img_dir+'_out'
        if msk_dir is not None and msk_outdir is None:
            msk_outdir = msk_dir+'_out'

        self.img_outdir=img_outdir 
        self.msk_outdir=msk_outdir

        # create output directory if needed
        if not os.path.exists(self.img_outdir):
            os.makedirs(self.img_outdir, exist_ok=True)
        if msk_dir is not None and  not os.path.exists(self.msk_outdir):
            os.makedirs(self.msk_outdir, exist_ok=True)

        self.num_classes = num_classes

        self.remove_bg = remove_bg

        self.median_spacing = np.array(median_spacing)
        self.clipping_bounds = np.array(clipping_bounds)
        self.intensity_moments = intensity_moments
        self.use_tif = use_tif

        # self.one_hot = tio.OneHot()
    
    def prepare(self):
        print("preprocessing...")
        for i in tqdm(range(len(self.img_fnames))):
            # set image and mask name
            img_fname = self.img_fnames[i]
            img_path = os.path.join(self.img_dir, img_fname)
            if self.msk_dir: msk_path = os.path.join(self.msk_dir, img_fname)
            # read image and mask
            img,spacing = adaptive_imread(img_path)
            if self.msk_dir: msk,_ = adaptive_imread(msk_path)

            # extend dim
            img = np.expand_dims(img, 0)

            # one hot encoding for the mask
            if self.msk_dir: 
                msk = one_hot(msk, self.num_classes)
                if self.remove_bg:
                    msk = msk[1:]


            # TODO: remove torchio dependency! 
            # resample the image if needed
            # if len(self.median_spacing)>0:
            #     resample = (self.median_spacing/spacing)[::-1] # transpose the dimension
            #     if resample.sum() > 0.1: # otherwise no need of spacing 
            #         if self.msk_dir: 
            #             sub = tio.Subject(img=tio.ScalarImage(tensor=img), msk=tio.LabelMap(tensor=msk))
            #             sub = tio.Resample(resample)(sub)
            #             img, msk = sub.img.numpy(), sub.msk.numpy()
            #         else:
            #             img = tio.Resample(resample)(img)

            # clip img
            if len(self.clipping_bounds)>0:
                img = np.clip(img, self.clipping_bounds[0], self.clipping_bounds[1])

            # normalize the image and msk
            # z-score normalization for the image
            if len(self.intensity_moments)>0:
                img = (img-self.intensity_moments[0])/self.intensity_moments[1]
            else:
                img = (img-img.mean())/img.std()
            
            # enhance contrast
            # img = exposure.equalize_hist(img)

            # range image in [-1, 1]
            # img = (img - img.min())/(img.max()-img.min()) * 2 - 1

            # set image type
            img = img.astype(np.float32)
            if self.msk_dir: msk = msk.astype(np.byte)

            # save the image and the mask as tif
            img_fname = os.path.basename(img_path).split('.')[0]
            # save image
            img_out_path = os.path.join(self.img_outdir, img_fname+'.tif')
            imsave(img_out_path, img)
            # save mask
            if self.msk_outdir: 
                msk_out_path = os.path.join(self.msk_outdir, img_fname+'.tif')
                imsave(msk_out_path, msk)
        print("done preprocessing!")

def preprocess(
    img_dir,
    msk_dir,
    img_outdir,
    msk_outdir,
    num_classes,
    remove_bg=True,
):
    Preprocessing(
        img_dir=img_dir,
        msk_dir=msk_dir,
        img_outdir=img_outdir,
        msk_outdir=msk_outdir,
        num_classes=num_classes,

        remove_bg=remove_bg,
        # median_spacing=[0.79492199, 0.79492199, 2.5],
        # clipping_bounds=[-109.0, 232.0],
        # intensity_moments=[69.6876,93.93239],
        use_tif=True,
    ).prepare()

#---------------------------------------------------------------------------

if __name__=='__main__':
    valid_names = {
        "preprocess": preprocess,
    }

    parser = argparse.ArgumentParser(description="Dataset preprocessing for training purpose.")
    parser.add_argument("-n", "--name", type=str, default="preprocess",
        help="Name of the tested method. Valid names: {}".format(list(valid_names.keys())))
    parser.add_argument("--img_dir", type=str,
        help="Path of the images directory")
    parser.add_argument("--msk_dir", type=str,
        help="Path to the masks/labels directory")
    parser.add_argument("--img_outdir", type=str,
        help="Path to the directory of the preprocessed images")
    parser.add_argument("--msk_outdir", type=str,
        help="Path to the directory of the preprocessed masks/labels")
    parser.add_argument("--num_classes", type=int, default=1,
        help="Number of classes (types of objects) in the dataset. The background is not included. (default=1)")
    parser.add_argument("--auto_config", default=False,  action='store_true', dest='auto_config',
        help="show the information to copy and paste inside the configuration file (patch_size, batch_size and num_pools).") 
    args = parser.parse_args()

    valid_names[args.name](
        img_dir=args.img_dir,
        msk_dir=args.msk_dir,
        img_outdir=args.img_outdir,
        msk_outdir=args.msk_outdir,
        num_classes=args.num_classes+1, # +1 for the background
    )

    if args.auto_config:
        import auto_config
        median = auto_config.compute_median(path=args.img_outdir)
        patch, pool, batch = auto_config.find_patch_pool_batch(dims=median, max_dims=(128,128,128))
        auto_config.display_info(patch, pool, batch)

#---------------------------------------------------------------------------




#---------------------------------------------------------------------------
# Method builder
# The main purpose of this class is to easily reload a training 
#---------------------------------------------------------------------------

import os
from skimage.io import imsave
import torch
from torch.utils.data import DataLoader
# from monai.data import ThreadDataLoader
import pandas as pd
# import torch.distributed as dist

from biom3d import register
from biom3d import callbacks as clbk
from biom3d import utils

#---------------------------------------------------------------------------
# utils to read config's functions in the function register

def read_config(config_fct, register_cat, **kwargs):
    """
    read the register category at register_cat and run the corresponding function 
    named in the config file
    """
    # read the corresponding name in the register
    register_fct_ = register_cat[config_fct.fct]

    # get the register function name
    register_fct = register_fct_.fct

    # get the register function kwargs and merge them with config function kwargs 
    register_kwargs = {**register_fct_.kwargs, **config_fct.kwargs, **kwargs} 

    # run the function with its kwargs
    return register_fct(**register_kwargs) 


#---------------------------------------------------------------------------
# for distributed data parallel

# TODO: use DDP...
# def setup_ddp(rank, world_size):
#     os.environ['MASTER_ADDR'] = 'localhost'
#     os.environ['MASTER_PORT'] = '12355'

#     # initialize the process group
#     dist.init_process_group("gloo", rank=rank, world_size=world_size)

# def cleanup_ddp():
#     dist.destroy_process_group()

#---------------------------------------------------------------------------
# self-supervised specific
# optimizer and parameters getter

def get_params_groups(model):
    regularized = []
    not_regularized = []
    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue
        # we do not regularize biases nor Norm parameters
        if name.endswith(".bias") or len(param.shape) == 1:
            not_regularized.append(param)
        else:
            regularized.append(param)
    return [{'params': regularized}, {'params': not_regularized, 'weight_decay': 0.}]

class LARS(torch.optim.Optimizer):
    """
    Almost copy-paste from https://github.com/facebookresearch/barlowtwins/blob/main/main.py
    """
    def __init__(self, params, lr=0, weight_decay=0, momentum=0.9, eta=0.001,
                 weight_decay_filter=None, lars_adaptation_filter=None):
        defaults = dict(lr=lr, weight_decay=weight_decay, momentum=momentum,
                        eta=eta, weight_decay_filter=weight_decay_filter,
                        lars_adaptation_filter=lars_adaptation_filter)
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(self):
        for g in self.param_groups:
            for p in g['params']:
                dp = p.grad

                if dp is None:
                    continue

                if p.ndim != 1:
                    dp = dp.add(p, alpha=g['weight_decay'])

                if p.ndim != 1:
                    param_norm = torch.norm(p)
                    update_norm = torch.norm(dp)
                    one = torch.ones_like(param_norm)
                    q = torch.where(param_norm > 0.,
                                    torch.where(update_norm > 0,
                                                (g['eta'] * param_norm / update_norm), one), one)
                    dp = dp.mul(q)

                param_state = self.state[p]
                if 'mu' not in param_state:
                    param_state['mu'] = torch.zeros_like(p)
                mu = param_state['mu']
                mu.mul_(g['momentum']).add_(dp)

                p.add_(mu, alpha=-g['lr'])

#---------------------------------------------------------------------------
# Global routine

class Builder:
    """
    the original goal of the builder was to easily save and load a model training.
    """
    def __init__(self, 
        config=None,         # inherit from Config class, stores the global variables
        path=None,      # path to a training folder
        training=True,  # use training mode or testing?
        ):                

        if path is not None:
            self.config = utils.load_config(path + "/log/config.yaml")
            # print(self.config)
            if training:
                self.load_train(path)
            else:
                self.load_test(path)
        else:
            assert config is not None, "[Error] config file not defined."
            self.config = config

            # if there are more than 1 GPU we augment the batch size and reduce the number of epochs
            if torch.cuda.device_count() > 1: 
                print("Let's use", torch.cuda.device_count(), "GPUs!")
                # Loop through all key-value pairs of a nested dictionary and change the batch_size 
                # for pairs in utils.nested_dict_pairs_iterator(self.config):
                #     if 'batch_size' in pairs:
                #         save = self.config[pairs[0]]; i=1
                #         while i < len(pairs) and pairs[i]!='batch_size':
                #             save = save[pairs[i]]; i+=1
                #         save['batch_size'] = torch.cuda.device_count()*self.config.BATCH_SIZE
                self.config = utils.nested_dict_change_value(self.config, 'batch_size', torch.cuda.device_count()*self.config.BATCH_SIZE)
                self.config.BATCH_SIZE *= torch.cuda.device_count()
                self.config.NB_EPOCHS = self.config.NB_EPOCHS//torch.cuda.device_count()

            # print(self.config)
            self.build_train()

    def build_dataset(self):
        """
        build the dataset.
        """
        if 'TRAIN_DATASET' in self.config.keys():
            self.train_dataset = read_config(self.config.TRAIN_DATASET, register.datasets)
        if 'TRAIN_DATALOADER' in self.config.keys():
            self.train_dataloader = read_config(self.config.TRAIN_DATALOADER, register.datasets)
        else:
            self.train_dataloader = DataLoader(self.train_dataset,**self.config.TRAIN_DATALOADER_KWARGS)

        if 'VAL_DATASET' in self.config.keys():
            self.val_dataset = read_config(self.config.VAL_DATASET, register.datasets)
        if 'VAL_DATALOADER' in self.config.keys():
            self.val_dataloader = read_config(self.config.VAL_DATALOADER, register.datasets)
        elif 'VAL_DATALOADER_KWARGS' in self.config.keys():
            self.val_dataloader = DataLoader(self.val_dataset,**self.config.VAL_DATALOADER_KWARGS)


    def build_model(self, training=True):
        """
        build the model, the losses and the metrics.
        """
        
        # self-supervised models are special cases
        # 2 models must be defined: student and teacher model
        if 'Self' in self.config.MODEL.fct:
            student = read_config(self.config.MODEL, register.models)
            teacher = read_config(self.config.MODEL, register.models)

            # if torch.cuda.is_available():
            student.cuda()
            teacher.cuda()
            
            # teacher and student start with the same weights
            teacher.load_state_dict(student.state_dict())

            # there is no backpropagation through the teacher, so no need for gradients
            for p in teacher.parameters():
                p.requires_grad = False

            # student model is DataParallel
            if torch.cuda.device_count() > 1:
                print("Let's use", torch.cuda.device_count(), "GPUs!")
                student = torch.nn.DataParallel(student)
            
            # verbose
            print("Student model:")
            print(student)

            # gather models
            self.model = [student, teacher]
        else: 
            self.model = read_config(self.config.MODEL, register.models)


            if torch.cuda.device_count() > 1:
                print("Let's use", torch.cuda.device_count(), "GPUs!")
                self.model = torch.nn.DataParallel(self.model)

            if torch.cuda.is_available():
                self.model.cuda()

            # TODO: use DDP...
            # if rank is not None: 
            #     self.model = torch.nn.parallel.DistributedDataParallel(self.model, device_ids=[rank])
            # else:
            #     self.model.cuda()
            # if training:
            #     self.model.train()
            # else:
            #     self.model.eval()

            print(self.model) # TODO: model verbose

        # losses
        if training:
            # self.loss_fn = DiceBCE(name='train_loss') # normal supervision
            self.loss_fn = read_config(self.config.TRAIN_LOSS, register.metrics)
            self.loss_fn.cuda().train()
            # self.loss_fn.train()

            if 'VAL_LOSS' in self.config.keys():
                self.val_loss_fn = read_config(self.config.VAL_LOSS, register.metrics)
                self.val_loss_fn.cuda().eval()
                # self.val_loss_fn.eval()
            else: 
                self.val_loss_fn = None

            # metrics
            if 'TRAIN_METRICS' in self.config.keys():
                self.train_metrics = [read_config(v, register.metrics) for v in self.config.TRAIN_METRICS.values()]
                # print(self.train_metrics)
                for m in self.train_metrics: m.cuda().eval()
            else: self.train_metrics = []

            if 'VAL_METRICS' in self.config.keys():
                self.val_metrics = [read_config(v, register.metrics) for v in self.config.VAL_METRICS.values()]
                for m in self.val_metrics: m.cuda().eval()
            else: self.val_metrics = []

            # optimizer
            if 'LR_START' in self.config.keys() and self.config.LR_START is not None: lr = self.config.LR_START
            else: lr = 1e-2
            print("Training will start with a learning rate of", lr)
            
            # for self-supervised learning:
            if 'Self' in self.config.MODEL.fct: params = get_params_groups(self.model[0])
            else: params = [{'params': self.model.parameters()}, {'params': self.loss_fn.parameters()}]

            weight_decay = 0 if not ('WEIGHT_DECAY' in self.config.keys()) else 3e-5

            self.optim = torch.optim.SGD(
                # [{'params': self.model.parameters()}, {'params': self.loss_fn.parameters()}], 
                params, 
                lr=lr, momentum=0.99, nesterov=True, weight_decay=weight_decay)
                # lr=lr, momentum=0.99, nesterov=True)

            # self.optim = LARS(params)


    def build_callbacks(self):
        """
        build the callbacks for the training process.
        """

        clbk_dict = {}

        # callbacks: saver (model, summary, images, train_state), logger (prints), scheduler, model_freezer, telegram_sender
        if "LR_MILESTONES" in self.config.keys():
            self.clbk_scheduler = clbk.LRSchedulerMultiStep(
                optimizer=self.optim,
                milestones=self.config.LR_MILESTONES)
            clbk_dict["lr_scheduler"] = self.clbk_scheduler

        elif "LR_T_MAX" in self.config.keys():
            self.clbk_scheduler = clbk.LRSchedulerCosine(
                optimizer=self.optim,
                T_max=self.config.LR_T_MAX)
            clbk_dict["lr_scheduler"] = self.clbk_scheduler

        elif "LR_START" in self.config.keys():
            self.clbk_scheduler = clbk.LRSchedulerPoly(
                optimizer=self.optim,
                initial_lr=self.config.LR_START,
                max_epochs=self.config.NB_EPOCHS,
                )
            clbk_dict["lr_scheduler"] = self.clbk_scheduler

        if 'USE_FG_CLBK' in self.config.keys() and self.config.USE_FG_CLBK:
            self.clbk_fg_scheduler = clbk.ForceFGScheduler(
                dataloader=self.train_dataloader,
                initial_rate=1,
                min_rate=0.3,
                max_epochs=self.config.NB_EPOCHS,
            )
            clbk_dict["fg_scheduler"] = self.clbk_fg_scheduler
        
        if 'USE_WD_CLBK' in self.config.keys() and self.config.USE_WD_CLBK:
            self.clbk_wd_scheduler = clbk.WeightDecayScheduler(
                optimizer=self.optim,
                initial_wd=self.config.INITIAL_WD,
                final_wd=self.config.FINAL_WD,
                nb_epochs=self.config.NB_EPOCHS,
                use_poly=True,
            )
            clbk_dict["wd_scheduler"] = self.clbk_wd_scheduler
        
        if 'USE_MOMENTUM_CLBK' in self.config.keys() and self.config.USE_MOMENTUM_CLBK:
            self.clbk_momentum_scheduler = clbk.MomentumScheduler(
                initial_momentum=0.9,
                final_momentum=self.config.INITIAL_MOMENTUM,
                nb_epochs=50,
                mode='linear',
                
                # initial_momentum=self.config.INITIAL_MOMENTUM,
                # final_momentum=1.0,
                # nb_epochs=self.config.NB_EPOCHS,
                # mode='exp',
                # exponent=6.0,
            )
            clbk_dict["momentum_scheduler"] = self.clbk_momentum_scheduler
        
        if 'USE_OVERLAP_CLBK' in self.config.keys() and self.config.USE_OVERLAP_CLBK:
            self.clbk_overlap_scheduler = clbk.OverlapScheduler(
                dataloader=self.train_dataloader,
                initial_rate=1.0,
                min_rate=-1.0,
                max_epochs=self.config.NB_EPOCHS,
            )
            clbk_dict["overlap_scheduler"] = self.clbk_overlap_scheduler
        
        if 'USE_GLOBAL_SCALE_CLBK' in self.config.keys() and self.config.USE_GLOBAL_SCALE_CLBK:
            self.clbk_global_scale_scheduler = clbk.GlobalScaleScheduler(
                dataloader=self.train_dataloader,
                initial_rate=1.0,
                min_rate=0.0,
                max_epochs=self.config.NB_EPOCHS,
            )
            clbk_dict["clbk_global_scale_scheduler"] = self.clbk_global_scale_scheduler

        if 'USE_DATASET_SIZE_CLBK' in self.config.keys() and self.config.USE_DATASET_SIZE_CLBK:
            self.clbk_dataset_size_scheduler = clbk.DatasetSizeScheduler(
                dataloader=self.train_dataloader,
                model=self.model,
                max_dataset_size=len(self.train_dataloader.dataset.train_imgs),
                min_dataset_size=5,
            )
            clbk_dict["clbk_dataset_size_scheduler"] = self.clbk_dataset_size_scheduler
            
        self.clbk_modelsaver = clbk.ModelSaver(
                model=self.model,
                optimizer=self.optim, 
                path=self.model_path, 
                every_epoch=self.config.SAVE_MODEL_EVERY_EPOCH, 
                save_best=self.config.SAVE_BEST, 
                loss=self.val_loss_fn if self.val_loss_fn else self.loss_fn,
                saved_loss=self.loss_fn)
        clbk_dict["model_saver"] = self.clbk_modelsaver

        # self.clbk_logsaver = clbk.LogSaver(
        #         log_dir=self.log_dir,
        #         train_loss=self.loss_fn,
        #         val_loss=self.val_loss_fn,
        #         train_metrics=self.train_metrics,
        #         val_metrics=self.val_metrics,
        #         scheduler=self.clbk_scheduler,
        #         save_best=self.config.SAVE_BEST,
        #         every_batch=10)
        
        if "USE_IMAGE_CLBK" in self.config.keys() and self.config.USE_IMAGE_CLBK and hasattr(self, 'val_dataloader'):
            self.clbk_imagesaver = clbk.ImageSaver(
                    image_dir=self.image_dir,
                    model=self.model,
                    val_dataloader=self.val_dataloader,
                    every_epoch=self.config.SAVE_IMAGE_EVERY_EPOCH,
                    use_sigmoid=not self.config.USE_SOFTMAX,
                    )
            clbk_dict["image_saver"] = self.clbk_imagesaver

        
        self.clbk_tensorboardsaver = clbk.TensorboardSaver(
                log_dir=self.log_dir,
                train_loss=self.loss_fn,
                val_loss=self.val_loss_fn,
                train_metrics=self.train_metrics,
                val_metrics=self.val_metrics,
                batch_size=self.config.BATCH_SIZE,
                n_batch_per_epoch=len(self.train_dataloader))
        clbk_dict["tensorboard_saver"] = self.clbk_tensorboardsaver
        
        self.clbk_logprinter = clbk.LogPrinter(
                metrics=[self.loss_fn] + self.train_metrics if self.train_metrics else [self.loss_fn], 
                nbof_epochs=self.config.NB_EPOCHS, 
                nbof_batches=len(self.train_dataloader),
                every_batch=10)
        clbk_dict["log_printer"] = self.clbk_logprinter
        
        # [DEPRECATED]
        # self.clbk_telegram = clbk.Telegram(
        #         loss=self.loss_fn,
        #         test_loss=self.val_loss_fn)
        # clbk_list += [self.clbk_telegram]
        
        self.clbk_metricupdater = clbk.MetricsUpdater(
                metrics=[self.loss_fn] + self.train_metrics if self.train_metrics else [self.loss_fn], 
                batch_size=self.config.BATCH_SIZE)
        clbk_dict["metric_updater"] = self.clbk_metricupdater

        self.callbacks = clbk.Callbacks(clbk_dict)


    def build_train(self):
        """
        build the method from scratch.
        """
        # saver configs
        self.base_dir, self.image_dir, self.log_dir, self.model_dir = utils.create_save_dirs(
            self.config.LOG_DIR, self.config.DESC, dir_names=['image', 'log', 'model'], return_base_dir=True) 
        utils.save_config(os.path.join(self.log_dir, 'config.yaml'), self.config) # save the config file

        self.model_path = os.path.join(self.model_dir, self.config.DESC)

        # first epoch
        self.initial_epoch = 0

        # Build the method
        self.build_dataset()
        self.build_model()
        self.build_callbacks()

    def run_training(self):
        """
        run the training and validation routines.
        """
        if 'USE_FP16' in self.config.keys() and self.config.USE_FP16:
            scaler = torch.cuda.amp.GradScaler()
        else:
            scaler = None
        self.callbacks.on_train_begin()
        for epoch in range(self.initial_epoch, self.config.NB_EPOCHS):
            self.callbacks.on_epoch_begin(epoch)
            read_config(
                self.config.TRAINER,
                register.trainers,
                scaler=scaler,
                dataloader=self.train_dataloader,
                model=self.model,
                loss_fn=self.loss_fn,
                optimizer=self.optim,
                callbacks=self.callbacks,
                metrics=self.train_metrics,
                epoch=epoch,
                use_deep_supervision=self.config.USE_DEEP_SUPERVISION,
                )
            if (epoch % self.config.VAL_EVERY_EPOCH == 0) or (epoch == (self.config.NB_EPOCHS-1)):
                if 'VALIDATER' in self.config.keys():
                    read_config(
                        self.config.VALIDATER,
                        register.trainers,
                        dataloader=self.val_dataloader,
                        model=self.model,
                        loss_fn=self.val_loss_fn,
                        metrics=self.val_metrics,
                        use_fp16=scaler is not None)
            self.callbacks.on_epoch_end(epoch)
        self.callbacks.on_train_end()
    
    # def main_ddp(self, rank, world_size): # TODO: use DDP...
    #     setup_ddp(rank, world_size)

    #     self.build_model()

    #     self.run_training()
    #     cleanup_ddp()

    # def run_training_ddp(self):
    #     torch.multiprocessing.spawn(
    #         self.main_ddp,
    #         args=(torch.cuda.device_count(),),
    #         nprocs=torch.cuda.device_count(),
    #         join=True,
    #     )

    def run_prediction_single(self, img_path, return_logit=True):
        """
        TODO: this function should be defined in another file 
        compute a prediction for one image, just for visualizing
        """

        return read_config(
            self.config.PREDICTOR, 
            register.predictors,
            img_path = img_path,
            model = self.model,
            return_logit = return_logit,
            )
    
    def run_prediction_folder(self, dir_in, dir_out, return_logit=False):
        """
        compute a prediction for one image, just for visualizing
        """
        fnames_in = sorted(utils.abs_listdir(dir_in))
        if not os.path.isdir(dir_out):
            os.makedirs(dir_out, exist_ok=True)
        
        # remove extension
        fnames_out = [f[:f.rfind('.')] for f in sorted(os.listdir(dir_in))]
        # fnames_out = [os.path.basename(f).split('.')[0] for f in sorted(os.listdir(dir_in))]

        # add folder path
        fnames_out = utils.abs_path(dir_out,fnames_out)
        

        for i, img_path in enumerate(fnames_in):
            print("running prediction for image: ", img_path)
            pred = self.run_prediction_single(img_path=img_path, return_logit=return_logit)
            print("Saving images in", fnames_out[i]+".tif")
            imsave(fnames_out[i]+".tif", pred)
    
    def load_train(self, 
        path, 
        load_best=False): # whether to load the best model
        """
        load a builder from a folder.
        the folder should have been created by the self.run_training method.
        """

        # setup the different paths from the folder
        self.base_dir = path
        self.image_dir = os.path.join(path, 'image')
        self.log_dir = os.path.join(path, 'log')
        self.model_dir = os.path.join(path, 'model')

        # self.log_path = os.path.join(self.log_dir, 'log.csv')
        # self.log_best_path = os.path.join(self.log_dir, 'log_best.csv')

        self.model_path = os.path.join(self.model_dir, self.config.DESC)

        # call the build method
        self.build_model()

        # load the model and the optimizer and the loss if needed
        model_name_full = self.config.DESC + '_best.pth' if load_best else self.config.DESC + '.pth'
        ckpt = torch.load(os.path.join(self.model_dir, model_name_full))
        print("Loading model from", os.path.join(self.model_dir, model_name_full))
        print(self.model.load_state_dict(ckpt['model'], strict=False))
        if 'loss' in ckpt.keys(): self.loss_fn.load_state_dict(ckpt['loss'])

        if not 'LR_START' in self.config.keys() or self.config.LR_START is None:
            self.optim.load_state_dict(ckpt['opt'])

        if 'epoch' in list(ckpt.keys()): # tmp
            self.initial_epoch=ckpt['epoch'] # definitive version 
        # else: # tmp
        #     with open(self.log_path, "r") as file:
        #         last_line = file.readlines()[-1] # read the last line
        #     last_line = last_line.split(',')
        #     self.initial_epoch=int(last_line[0])
        print('Restart training at epoch {}'.format(self.initial_epoch))

        self.build_dataset()

        # load callbacks
        self.build_callbacks()

        # load the best loss for the model saver and the log saver
        # if os.path.exists(self.log_best_path):
        #     df_best = pd.read_csv(self.log_best_path)
        #     if hasattr(self, 'clbk_logsaver'):
        #         self.clbk_logsaver.best_loss= df_best.val_loss[0]
        #         print('Loads log saver')
        #     if hasattr(self, 'clbk_modelsaver'):
        #         self.clbk_modelsaver.best_loss= df_best.val_loss[0]
        #         print('Load model saver')
    
    def load_test(self, 
        path, 
        load_best=True): # whether to load the best model
        """
        load a builder from a folder.
        the folder should have been created by the self.run_training method.
        """

        # setup the different paths from the folder
        self.model_dir = os.path.join(path, 'model')
        self.model_path = os.path.join(self.model_dir, self.config.DESC)

        # call the build method
        self.build_model(training=False)

        # load the model and the optimizer
        model_name_full = self.config.DESC + '_best.pth' if load_best else self.config.DESC + '.pth'
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        ckpt = torch.load(os.path.join(self.model_dir, model_name_full), map_location=torch.device(device))
        print("Loading model from", os.path.join(self.model_dir, model_name_full))

        # remove `module.` prefix
        state_dict = {k.replace("module.", ""): v for k, v in ckpt['model'].items()}  
        # remove `backbone.` prefix induced by multicrop wrapper
        state_dict = {k.replace("backbone.", ""): v for k, v in state_dict.items()}
        print(self.model.load_state_dict(state_dict, strict=False))

#---------------------------------------------------------------------------